# Generated from WBT.g4 by ANTLR 4.11.1
# encoding: utf-8
from io import StringIO
import sys

from antlr4 import *

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO

def serializedATN():
    return [
        4,1,15,111,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,1,0,1,0,5,0,25,8,0,10,0,12,0,
        28,9,0,1,1,1,1,1,1,1,1,1,2,1,2,1,2,1,2,1,2,1,2,3,2,40,8,2,1,3,1,
        3,1,3,5,3,45,8,3,10,3,12,3,48,9,3,1,4,1,4,1,4,1,4,1,4,3,4,55,8,4,
        1,5,1,5,1,5,1,5,1,5,3,5,62,8,5,1,6,4,6,65,8,6,11,6,12,6,66,1,7,1,
        7,1,8,1,8,1,8,5,8,74,8,8,10,8,12,8,77,9,8,1,8,3,8,80,8,8,1,8,1,8,
        1,8,1,8,1,8,4,8,87,8,8,11,8,12,8,88,1,8,3,8,92,8,8,1,8,1,8,3,8,96,
        8,8,1,9,3,9,99,8,9,1,9,1,9,3,9,103,8,9,4,9,105,8,9,11,9,12,9,106,
        1,10,1,10,1,10,0,0,11,0,2,4,6,8,10,12,14,16,18,20,0,1,1,0,8,9,117,
        0,26,1,0,0,0,2,29,1,0,0,0,4,33,1,0,0,0,6,46,1,0,0,0,8,54,1,0,0,0,
        10,61,1,0,0,0,12,64,1,0,0,0,14,68,1,0,0,0,16,95,1,0,0,0,18,104,1,
        0,0,0,20,108,1,0,0,0,22,25,3,4,2,0,23,25,3,2,1,0,24,22,1,0,0,0,24,
        23,1,0,0,0,25,28,1,0,0,0,26,24,1,0,0,0,26,27,1,0,0,0,27,1,1,0,0,
        0,28,26,1,0,0,0,29,30,5,1,0,0,30,31,5,12,0,0,31,32,3,4,2,0,32,3,
        1,0,0,0,33,34,5,12,0,0,34,35,5,2,0,0,35,36,5,15,0,0,36,37,3,6,3,
        0,37,39,5,3,0,0,38,40,5,15,0,0,39,38,1,0,0,0,39,40,1,0,0,0,40,5,
        1,0,0,0,41,42,3,8,4,0,42,43,5,15,0,0,43,45,1,0,0,0,44,41,1,0,0,0,
        45,48,1,0,0,0,46,44,1,0,0,0,46,47,1,0,0,0,47,7,1,0,0,0,48,46,1,0,
        0,0,49,50,5,4,0,0,50,51,5,12,0,0,51,55,3,10,5,0,52,53,5,12,0,0,53,
        55,3,10,5,0,54,49,1,0,0,0,54,52,1,0,0,0,55,9,1,0,0,0,56,62,3,12,
        6,0,57,62,3,14,7,0,58,62,3,16,8,0,59,62,3,4,2,0,60,62,3,20,10,0,
        61,56,1,0,0,0,61,57,1,0,0,0,61,58,1,0,0,0,61,59,1,0,0,0,61,60,1,
        0,0,0,62,11,1,0,0,0,63,65,5,13,0,0,64,63,1,0,0,0,65,66,1,0,0,0,66,
        64,1,0,0,0,66,67,1,0,0,0,67,13,1,0,0,0,68,69,5,14,0,0,69,15,1,0,
        0,0,70,75,5,5,0,0,71,72,5,15,0,0,72,74,3,10,5,0,73,71,1,0,0,0,74,
        77,1,0,0,0,75,73,1,0,0,0,75,76,1,0,0,0,76,79,1,0,0,0,77,75,1,0,0,
        0,78,80,5,15,0,0,79,78,1,0,0,0,79,80,1,0,0,0,80,81,1,0,0,0,81,96,
        5,6,0,0,82,83,5,5,0,0,83,86,3,18,9,0,84,85,5,7,0,0,85,87,3,18,9,
        0,86,84,1,0,0,0,87,88,1,0,0,0,88,86,1,0,0,0,88,89,1,0,0,0,89,91,
        1,0,0,0,90,92,5,7,0,0,91,90,1,0,0,0,91,92,1,0,0,0,92,93,1,0,0,0,
        93,94,5,6,0,0,94,96,1,0,0,0,95,70,1,0,0,0,95,82,1,0,0,0,96,17,1,
        0,0,0,97,99,5,15,0,0,98,97,1,0,0,0,98,99,1,0,0,0,99,100,1,0,0,0,
        100,102,5,13,0,0,101,103,5,15,0,0,102,101,1,0,0,0,102,103,1,0,0,
        0,103,105,1,0,0,0,104,98,1,0,0,0,105,106,1,0,0,0,106,104,1,0,0,0,
        106,107,1,0,0,0,107,19,1,0,0,0,108,109,7,0,0,0,109,21,1,0,0,0,15,
        24,26,39,46,54,61,66,75,79,88,91,95,98,102,106
    ]

class WBTParser ( Parser ):

    grammarFileName = "WBT.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'DEF'", "' {'", "'}'", "'hidden'", "'['", 
                     "']'", "','", "'TRUE'", "'FALSE'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "Comment", "Whitespace", 
                      "Identifier", "Number", "String", "Newline" ]

    RULE_world = 0
    RULE_defn = 1
    RULE_node = 2
    RULE_nodeBody = 3
    RULE_attribute = 4
    RULE_value = 5
    RULE_vector = 6
    RULE_string = 7
    RULE_array = 8
    RULE_vectorWithNewlines = 9
    RULE_boolean = 10

    ruleNames =  [ "world", "defn", "node", "nodeBody", "attribute", "value", 
                   "vector", "string", "array", "vectorWithNewlines", "boolean" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    Comment=10
    Whitespace=11
    Identifier=12
    Number=13
    String=14
    Newline=15

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.11.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class WorldContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def node(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(WBTParser.NodeContext)
            else:
                return self.getTypedRuleContext(WBTParser.NodeContext,i)


        def defn(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(WBTParser.DefnContext)
            else:
                return self.getTypedRuleContext(WBTParser.DefnContext,i)


        def getRuleIndex(self):
            return WBTParser.RULE_world

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitWorld" ):
                return visitor.visitWorld(self)
            else:
                return visitor.visitChildren(self)




    def world(self):

        localctx = WBTParser.WorldContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_world)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 26
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1 or _la==12:
                self.state = 24
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [12]:
                    self.state = 22
                    self.node()
                    pass
                elif token in [1]:
                    self.state = 23
                    self.defn()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 28
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DefnContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Identifier(self):
            return self.getToken(WBTParser.Identifier, 0)

        def node(self):
            return self.getTypedRuleContext(WBTParser.NodeContext,0)


        def getRuleIndex(self):
            return WBTParser.RULE_defn

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDefn" ):
                return visitor.visitDefn(self)
            else:
                return visitor.visitChildren(self)




    def defn(self):

        localctx = WBTParser.DefnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_defn)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 29
            self.match(WBTParser.T__0)
            self.state = 30
            self.match(WBTParser.Identifier)
            self.state = 31
            self.node()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NodeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Identifier(self):
            return self.getToken(WBTParser.Identifier, 0)

        def Newline(self, i:int=None):
            if i is None:
                return self.getTokens(WBTParser.Newline)
            else:
                return self.getToken(WBTParser.Newline, i)

        def nodeBody(self):
            return self.getTypedRuleContext(WBTParser.NodeBodyContext,0)


        def getRuleIndex(self):
            return WBTParser.RULE_node

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNode" ):
                return visitor.visitNode(self)
            else:
                return visitor.visitChildren(self)




    def node(self):

        localctx = WBTParser.NodeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_node)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            self.match(WBTParser.Identifier)
            self.state = 34
            self.match(WBTParser.T__1)
            self.state = 35
            self.match(WBTParser.Newline)
            self.state = 36
            self.nodeBody()
            self.state = 37
            self.match(WBTParser.T__2)
            self.state = 39
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,2,self._ctx)
            if la_ == 1:
                self.state = 38
                self.match(WBTParser.Newline)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NodeBodyContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def attribute(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(WBTParser.AttributeContext)
            else:
                return self.getTypedRuleContext(WBTParser.AttributeContext,i)


        def Newline(self, i:int=None):
            if i is None:
                return self.getTokens(WBTParser.Newline)
            else:
                return self.getToken(WBTParser.Newline, i)

        def getRuleIndex(self):
            return WBTParser.RULE_nodeBody

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNodeBody" ):
                return visitor.visitNodeBody(self)
            else:
                return visitor.visitChildren(self)




    def nodeBody(self):

        localctx = WBTParser.NodeBodyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_nodeBody)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 46
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==4 or _la==12:
                self.state = 41
                self.attribute()
                self.state = 42
                self.match(WBTParser.Newline)
                self.state = 48
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AttributeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Identifier(self):
            return self.getToken(WBTParser.Identifier, 0)

        def value(self):
            return self.getTypedRuleContext(WBTParser.ValueContext,0)


        def getRuleIndex(self):
            return WBTParser.RULE_attribute

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAttribute" ):
                return visitor.visitAttribute(self)
            else:
                return visitor.visitChildren(self)




    def attribute(self):

        localctx = WBTParser.AttributeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_attribute)
        try:
            self.state = 54
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [4]:
                self.enterOuterAlt(localctx, 1)
                self.state = 49
                self.match(WBTParser.T__3)
                self.state = 50
                self.match(WBTParser.Identifier)
                self.state = 51
                self.value()
                pass
            elif token in [12]:
                self.enterOuterAlt(localctx, 2)
                self.state = 52
                self.match(WBTParser.Identifier)
                self.state = 53
                self.value()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def vector(self):
            return self.getTypedRuleContext(WBTParser.VectorContext,0)


        def string(self):
            return self.getTypedRuleContext(WBTParser.StringContext,0)


        def array(self):
            return self.getTypedRuleContext(WBTParser.ArrayContext,0)


        def node(self):
            return self.getTypedRuleContext(WBTParser.NodeContext,0)


        def boolean(self):
            return self.getTypedRuleContext(WBTParser.BooleanContext,0)


        def getRuleIndex(self):
            return WBTParser.RULE_value

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValue" ):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)




    def value(self):

        localctx = WBTParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_value)
        try:
            self.state = 61
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [13]:
                self.enterOuterAlt(localctx, 1)
                self.state = 56
                self.vector()
                pass
            elif token in [14]:
                self.enterOuterAlt(localctx, 2)
                self.state = 57
                self.string()
                pass
            elif token in [5]:
                self.enterOuterAlt(localctx, 3)
                self.state = 58
                self.array()
                pass
            elif token in [12]:
                self.enterOuterAlt(localctx, 4)
                self.state = 59
                self.node()
                pass
            elif token in [8, 9]:
                self.enterOuterAlt(localctx, 5)
                self.state = 60
                self.boolean()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VectorContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Number(self, i:int=None):
            if i is None:
                return self.getTokens(WBTParser.Number)
            else:
                return self.getToken(WBTParser.Number, i)

        def getRuleIndex(self):
            return WBTParser.RULE_vector

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVector" ):
                return visitor.visitVector(self)
            else:
                return visitor.visitChildren(self)




    def vector(self):

        localctx = WBTParser.VectorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_vector)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 64 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 63
                self.match(WBTParser.Number)
                self.state = 66 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==13):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class StringContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def String(self):
            return self.getToken(WBTParser.String, 0)

        def getRuleIndex(self):
            return WBTParser.RULE_string

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitString" ):
                return visitor.visitString(self)
            else:
                return visitor.visitChildren(self)




    def string(self):

        localctx = WBTParser.StringContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_string)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 68
            self.match(WBTParser.String)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrayContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Newline(self, i:int=None):
            if i is None:
                return self.getTokens(WBTParser.Newline)
            else:
                return self.getToken(WBTParser.Newline, i)

        def value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(WBTParser.ValueContext)
            else:
                return self.getTypedRuleContext(WBTParser.ValueContext,i)


        def vectorWithNewlines(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(WBTParser.VectorWithNewlinesContext)
            else:
                return self.getTypedRuleContext(WBTParser.VectorWithNewlinesContext,i)


        def getRuleIndex(self):
            return WBTParser.RULE_array

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArray" ):
                return visitor.visitArray(self)
            else:
                return visitor.visitChildren(self)




    def array(self):

        localctx = WBTParser.ArrayContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_array)
        self._la = 0 # Token type
        try:
            self.state = 95
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,11,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 70
                self.match(WBTParser.T__4)
                self.state = 75
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,7,self._ctx)
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt==1:
                        self.state = 71
                        self.match(WBTParser.Newline)
                        self.state = 72
                        self.value() 
                    self.state = 77
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,7,self._ctx)

                self.state = 79
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==15:
                    self.state = 78
                    self.match(WBTParser.Newline)


                self.state = 81
                self.match(WBTParser.T__5)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 82
                self.match(WBTParser.T__4)
                self.state = 83
                self.vectorWithNewlines()
                self.state = 86 
                self._errHandler.sync(self)
                _alt = 1
                while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 84
                        self.match(WBTParser.T__6)
                        self.state = 85
                        self.vectorWithNewlines()

                    else:
                        raise NoViableAltException(self)
                    self.state = 88 
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,9,self._ctx)

                self.state = 91
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==7:
                    self.state = 90
                    self.match(WBTParser.T__6)


                self.state = 93
                self.match(WBTParser.T__5)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VectorWithNewlinesContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Number(self, i:int=None):
            if i is None:
                return self.getTokens(WBTParser.Number)
            else:
                return self.getToken(WBTParser.Number, i)

        def Newline(self, i:int=None):
            if i is None:
                return self.getTokens(WBTParser.Newline)
            else:
                return self.getToken(WBTParser.Newline, i)

        def getRuleIndex(self):
            return WBTParser.RULE_vectorWithNewlines

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVectorWithNewlines" ):
                return visitor.visitVectorWithNewlines(self)
            else:
                return visitor.visitChildren(self)




    def vectorWithNewlines(self):

        localctx = WBTParser.VectorWithNewlinesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_vectorWithNewlines)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 104 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 98
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==15:
                    self.state = 97
                    self.match(WBTParser.Newline)


                self.state = 100
                self.match(WBTParser.Number)
                self.state = 102
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,13,self._ctx)
                if la_ == 1:
                    self.state = 101
                    self.match(WBTParser.Newline)


                self.state = 106 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==13 or _la==15):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class BooleanContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return WBTParser.RULE_boolean

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitBoolean" ):
                return visitor.visitBoolean(self)
            else:
                return visitor.visitChildren(self)




    def boolean(self):

        localctx = WBTParser.BooleanContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_boolean)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 108
            _la = self._input.LA(1)
            if not(_la==8 or _la==9):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








"""Objects representing distributions that can be sampled from."""

import collections
import functools
import itertools
import math
import numbers
import random
import typing
import warnings

import numpy

from scenic.core.errors import InvalidScenarioError, ScenicError
from scenic.core.lazy_eval import (
    LazilyEvaluable,
    dependencies,
    isLazy,
    makeDelayedFunctionCall,
    needsLazyEvaluation,
    needsSampling,
    requiredProperties,
    valueInContext,
)
from scenic.core.utils import (
    DefaultIdentityDict,
    argsToString,
    cached,
    get_type_hints,
    sqrt2,
)

## Misc


def supportInterval(thing):
    """Lower and upper bounds on this value, if known."""
    if hasattr(thing, "supportInterval"):
        return thing.supportInterval()
    elif isinstance(thing, (int, float)):
        return thing, thing
    else:
        return None, None


def supmin(*vals):
    return None if None in vals else min(vals)


def supmax(*vals):
    return None if None in vals else max(vals)


def unionOfSupports(supports):
    mins, maxes = zip(*supports)
    return supmin(*mins), supmax(*maxes)


def addSupports(sup1, sup2):
    l1, u1 = sup1
    l2, u2 = sup2
    lower = None if l1 is None or l2 is None else l1 + l2
    upper = None if u1 is None or u2 is None else u1 + u2
    return lower, upper


def underlyingFunction(thing):
    """Original function underlying a distribution wrapper."""
    func = getattr(thing, "__wrapped__", thing)
    return getattr(func, "__func__", func)


def canUnpackDistributions(func):
    """Whether the function supports iterable unpacking of distributions."""
    return getattr(func, "_canUnpackDistributions", False)


def unpacksDistributions(func):
    """Decorator indicating the function supports iterable unpacking of distributions."""
    func._canUnpackDistributions = True
    return func


class RejectionException(Exception):
    """Exception used to signal that the sample currently being generated must be rejected."""

    pass


class RandomControlFlowError(ScenicError):
    """Exception indicating illegal conditional control flow depending on a random value.

    This includes trying to iterate over a random value, making a range of random length, etc.
    """

    pass


## Abstract distributions


class Samplable(LazilyEvaluable):
    """Abstract class for values which can be sampled, possibly depending on other values.

    Samplables may specify a proxy object which must have the same distribution as the
    original after conditioning on the scenario's requirements. This allows transparent
    conditioning without modifying Samplable fields of immutable objects.

    Args:
        dependencies: sequence of values that this value may depend on (formally, objects
            for which sampled values must be provided to `sampleGiven`). It is legal to
            include values which are not instances of `Samplable`, e.g. integers.

    Attributes:
        _conditioned: proxy object as described above; set using `conditionTo`.
        _dependencies: tuple of other samplables which must be sampled before this one;
            set by the initializer and subsequently immutable.
    """

    def __init__(self, dependencies):
        deps = []
        props = set()
        for dep in dependencies:
            if isLazy(dep):
                deps.append(dep)
                props.update(requiredProperties(dep))
        super().__init__(props, deps)
        self._conditioned = self  # version (partially) conditioned on requirements

    @staticmethod
    def sampleAll(quantities):
        """Sample all the given Samplables, which may have dependencies in common.

        Reproducibility note: the order in which the quantities are given can affect the
        order in which calls to random are made, affecting the final result.
        """
        subsamples = DefaultIdentityDict()
        for q in quantities:
            if q not in subsamples:
                subsamples[q] = q.sample(subsamples) if needsSampling(q) else q
        return subsamples

    def sample(self, subsamples=None):
        """Sample this value, optionally given some values already sampled."""
        if subsamples is None:
            subsamples = DefaultIdentityDict()
        for child in self._conditioned._dependencies:
            if child not in subsamples:
                subsamples[child] = child.sample(subsamples)
        return self._conditioned.sampleGiven(subsamples)

    def sampleGiven(self, value):
        """Sample this value, given values for all its dependencies.

        Implemented by subclasses.

        Args:
            value (DefaultIdentityDict): dictionary mapping objects to their sampled
                values. Guaranteed to provide values for all objects given in the set of
                dependencies when this `Samplable` was created.
        """
        raise NotImplementedError

    def serializeValue(self, values, serializer):
        for child in self._conditioned._dependencies:
            serializer.writeSamplable(child, values)

    def deserializeValue(self, serializer, values):
        for child in self._conditioned._dependencies:
            serializer.readSamplable(child, values)
        return self._conditioned.sampleGiven(values)

    def conditionTo(self, value):
        """Condition this value to another value with the same conditional distribution."""
        assert isinstance(value, Samplable)
        self._conditioned = value

    def evaluateIn(self, context):
        """See `LazilyEvaluable.evaluateIn`."""
        value = super().evaluateIn(context)
        # Check that all dependencies have been evaluated
        assert all(not needsLazyEvaluation(dep) for dep in value._dependencies)
        return value


class ConstantSamplable(Samplable):
    """A samplable which always evaluates to a constant value.

    Only for internal use.
    """

    def __init__(self, value):
        assert not isLazy(value), value
        self.value = value
        super().__init__(())

    def sampleGiven(self, value):
        return self.value


class Distribution(Samplable):
    """Abstract class for distributions.

    .. note::

        When called during dynamic simulations (vs. scenario compilation), constructors
        for distributions return *actual sampled values*, not `Distribution` objects.

    Args:
        dependencies: values which this distribution may depend on (see `Samplable`).
        valueType: **_valueType** to use (see below), or `None` for the default.

    Attributes:
        _valueType: type of the values sampled from this distribution, or `object` if the
            type is not known.
    """

    #: Default valueType for distributions of this class, when not otherwise specified.
    _defaultValueType = object

    #: Whether this type of distribution is a deterministic function of its dependencies.
    #:
    #: For example, `Options` is implemented as deterministic by using an internal
    #: `DiscreteRange` to select which of its finitely-many options to choose from: the
    #: value of the `Options` is then completely determined by the value of the range and
    #: the values of each of the options. This simplifies serialization because these
    #: dependencies likely have simpler valueTypes than the `Options` itself (e.g. if we
    #: had a random choice between a list and a string, encoding the actual sampled value
    #: would require saving type information).
    _deterministic = False

    def __new__(cls, *args, **kwargs):
        dist = super().__new__(cls)
        # During a simulation, return a sample from the distribution immediately
        import scenic.syntax.veneer as veneer

        if veneer.simulationInProgress():
            dist.__init__(*args, **kwargs)
            sim = veneer.simulation()
            subsamples = DefaultIdentityDict()
            # If we're replaying a previous simulation, use the saved value; otherwise sample one
            if sim.replayCanContinue():
                value = sim.replaySampledValue(dist, subsamples)
            else:
                value = dist.sample(subsamples)
            # Save the value for future replay
            subsamples[dist] = value
            sim.recordSampledValue(dist, subsamples)
            return value
        elif veneer.evaluatingRequirement:
            raise InvalidScenarioError(
                "cannot define distributions inside a requirement"
                " (try moving them outside)"
            )
        else:
            return dist

    def __init__(self, *dependencies, valueType=None):
        super().__init__(dependencies)
        self._needsSampling = True
        self._isLazy = True
        if valueType is None:
            valueType = self._defaultValueType
        self._valueType = valueType

    def clone(self):
        """Construct an independent copy of this Distribution.

        Optionally implemented by subclasses.
        """
        raise NotImplementedError("clone() not supported by this distribution")

    @property
    @cached
    def isPrimitive(self):
        """Whether this is a primitive Distribution."""
        try:
            self.clone()
            return True
        except NotImplementedError:
            return False

    def serializeValue(self, values, serializer):
        """Serialize the sampled value of this distribution.

        This method is used internally by `Scenario.sceneToBytes` and related APIs.
        If you define a new subclass of `Distribution`, you probably don't need to
        override this method. If your distribution has an unusual **valueType** (i.e.
        not `float`, `int`, or `Vector`), see the documentation for `Serializer` for
        instructions on how to support serialization.
        """
        if self._deterministic:
            # It suffices to just serialize our dependencies; this is preferable to
            # saving our sampled value directly, which could be arbitrarily complex.
            super().serializeValue(values, serializer)
        else:
            # Need to save our sampled value, which can't be reconstructed from the
            # values of our dependencies. There is then no need to save the latter.
            serializer.writeValue(values[self], self._valueType)

    def deserializeValue(self, serializer, values):
        if self._deterministic:
            return super().deserializeValue(serializer, values)
        else:
            return serializer.readValue(self._valueType)

    def bucket(self, buckets=None):
        """Construct a bucketed approximation of this Distribution.

        Optionally implemented by subclasses.

        This function factors a given Distribution into a discrete distribution over
        buckets together with a distribution for each bucket. The argument *buckets*
        controls how many buckets the domain of the original Distribution is split into.
        Since the result is an independent distribution, the original must support
        `clone`.
        """
        raise NotImplementedError("bucket() not supported by this distribution")

    def supportInterval(self):
        """Compute lower and upper bounds on the value of this Distribution.

        By default returns :scenic:`(None, None)` indicating that no lower or upper bounds
        are known. Subclasses may override this method to provide more accurate results.
        """
        return None, None

    def __getattr__(self, name):
        if name.startswith("__") and name.endswith("__"):  # ignore special attributes
            return object.__getattribute__(self, name)
        return AttributeDistribution(name, self)

    def __call__(self, *args, **kwargs):
        return OperatorDistribution("__call__", self, args, kwargs)

    def __iter__(self):
        raise RandomControlFlowError(f"cannot iterate through a random value")

    def _comparisonError(self, other):
        raise RandomControlFlowError(
            "random values cannot be compared " "(and control flow cannot depend on them)"
        )

    __lt__ = _comparisonError
    __le__ = _comparisonError
    __gt__ = _comparisonError
    __ge__ = _comparisonError
    __eq__ = _comparisonError
    __ne__ = _comparisonError

    def __hash__(self):  # need to explicitly define since we overrode __eq__
        return id(self)

    def __bool__(self):
        raise RandomControlFlowError("control flow cannot depend on a random value")


## Derived distributions


class TupleDistribution(Distribution, collections.abc.Sequence):
    """Distributions over tuples (or namedtuples, or lists)."""

    _deterministic = True

    def __init__(self, *coordinates, builder=tuple):
        super().__init__(*coordinates)
        self.coordinates = coordinates
        self.builder = builder

    def __len__(self):
        return len(self.coordinates)

    def __getitem__(self, index):
        return self.coordinates[index]

    def __iter__(self):
        yield from self.coordinates

    def sampleGiven(self, value):
        return self.builder(value[coordinate] for coordinate in self.coordinates)

    def evaluateInner(self, context):
        coordinates = (valueInContext(coord, context) for coord in self.coordinates)
        return TupleDistribution(*coordinates, builder=self.builder)

    def __repr__(self):
        coords = ", ".join(repr(c) for c in self.coordinates)
        if self.builder is tuple:
            return f"({coords},)"
        elif self.builder is list:
            return f"[{coords}]"
        else:
            return f"TupleDistribution({coords}, builder={self.builder!r})"


class SliceDistribution(Distribution):
    """Distributions over `slice` objects."""

    _deterministic = True

    def __init__(self, start, stop, step):
        super().__init__(start, stop, step, valueType=slice)
        self.start = start
        self.stop = stop
        self.step = step

    def sampleGiven(self, value):
        return slice(value[self.start], value[self.stop], value[self.step])

    def evaluateInner(self, context):
        start = valueInContext(self.start, context)
        stop = valueInContext(self.stop, context)
        step = valueInContext(self.step, context)
        return SliceDistribution(start, stop, step)

    def __repr__(self):
        return f"slice({self.start!r}, {self.stop!r}, {self.step!r})"


def toDistribution(val):
    """Wrap Python data types with Distributions, if necessary.

    For example, tuples containing Samplables need to be converted into TupleDistributions
    in order to keep track of dependencies properly.
    """
    if isinstance(val, (tuple, list)):
        coords = [toDistribution(c) for c in val]
        if any(isLazy(c) for c in coords):
            if isinstance(val, tuple) and hasattr(val, "_fields"):  # namedtuple
                builder = type(val)._make
            else:
                builder = type(val)
            return TupleDistribution(*coords, builder=builder)
    elif isinstance(val, slice):
        attrs = (val.start, val.stop, val.step)
        if any(isLazy(a) for a in attrs):
            return SliceDistribution(*attrs)
    return val


class FunctionDistribution(Distribution):
    """Distribution resulting from passing distributions to a function"""

    _deterministic = True

    def __init__(self, func, args, kwargs, support=None, valueType=None):
        args = tuple(toDistribution(arg) for arg in args)
        kwargs = {name: toDistribution(arg) for name, arg in kwargs.items()}
        if valueType is None:
            valueType = get_type_hints(func).get("return")
        super().__init__(*args, *kwargs.values(), valueType=valueType)
        self.function = func
        self.arguments = args
        self.kwargs = kwargs
        self.support = support

    def sampleGiven(self, value):
        args = []
        for arg in self.arguments:
            if isinstance(arg, StarredDistribution):
                val = value[arg]
                try:
                    iter(val)
                except TypeError:  # TODO improve backtrace
                    raise TypeError(
                        f"'{type(val).__name__}' object on line {arg.lineno} "
                        "is not iterable"
                    ) from None
                args.extend(val)
            else:
                args.append(value[arg])
        kwargs = {name: value[arg] for name, arg in self.kwargs.items()}
        return self.function(*args, **kwargs)

    def evaluateInner(self, context):
        function = valueInContext(self.function, context)
        arguments = tuple(valueInContext(arg, context) for arg in self.arguments)
        kwargs = {name: valueInContext(arg, context) for name, arg in self.kwargs.items()}
        return FunctionDistribution(function, arguments, kwargs)

    def supportInterval(self):
        if self.support is None:
            return None, None
        subsupports = (supportInterval(arg) for arg in self.arguments)
        kwss = {name: supportInterval(arg) for name, arg in self.kwargs.items()}
        return self.support(*subsupports, **kwss)

    def __repr__(self):
        return f"{self.function.__name__}({argsToString(self.arguments, self.kwargs)})"


def distributionFunction(wrapped=None, *, support=None, valueType=None):
    """Decorator for wrapping a function so that it can take distributions as arguments.

    This decorator is mainly for internal use, and is not necessary when defining a
    function in a Scenic file. It is, however, needed when calling external functions
    which contain control flow or other operations that Scenic distribution objects
    (representing random values) do not support.
    """
    if wrapped is None:  # written without arguments as @distributionFunction
        return lambda wrapped: distributionFunction(
            wrapped, support=support, valueType=valueType
        )

    @unpacksDistributions
    @functools.wraps(wrapped)
    def helper(*args, **kwargs):
        args = tuple(toDistribution(arg) for arg in args)
        kwargs = {name: toDistribution(arg) for name, arg in kwargs.items()}
        if any(needsSampling(arg) for arg in itertools.chain(args, kwargs.values())):
            return FunctionDistribution(wrapped, args, kwargs, support, valueType)
        elif any(
            needsLazyEvaluation(arg) for arg in itertools.chain(args, kwargs.values())
        ):
            # recursively call this helper (not the original function), since the
            # delayed arguments may evaluate to distributions, in which case we'll
            # have to make a FunctionDistribution
            return makeDelayedFunctionCall(helper, args, kwargs)
        else:
            return wrapped(*args, **kwargs)

    return helper


def monotonicDistributionFunction(method, valueType=None):
    """Like distributionFunction, but additionally specifies that the function is monotonic."""

    def support(*subsupports, **kwss):
        mins, maxes = zip(*subsupports)
        kwmins = {name: interval[0] for name, interval in kwss.items()}
        kwmaxes = {name: interval[1] for name, interval in kwss.items()}
        l = None if None in mins or None in kwmins else method(*mins, **kwmins)
        r = None if None in maxes or None in kwmaxes else method(*maxes, **kwmaxes)
        return l, r

    return distributionFunction(method, support=support, valueType=valueType)


class StarredDistribution(Distribution):
    """A placeholder for the iterable unpacking operator * applied to a distribution."""

    _deterministic = True

    def __init__(self, value, lineno):
        assert isinstance(value, Distribution)
        self.value = value
        self.lineno = lineno  # for error handling when unpacking fails
        super().__init__(value, valueType=value._valueType)

    def sampleGiven(self, value):
        return value[self.value]

    def evaluateInner(self, context):
        return StarredDistribution(valueInContext(self.value, context), self.lineno)

    def __repr__(self):
        return f"*{self.value!r}"


class MethodDistribution(Distribution):
    """Distribution resulting from passing distributions to a method of a fixed object"""

    _deterministic = True

    def __init__(self, method, obj, args, kwargs, valueType=None):
        args = tuple(toDistribution(arg) for arg in args)
        kwargs = {name: toDistribution(arg) for name, arg in kwargs.items()}
        if valueType is None:
            valueType = get_type_hints(method).get("return")
        super().__init__(*args, *kwargs.values(), valueType=valueType)
        self.method = method
        self.object = obj
        self.arguments = args
        self.kwargs = kwargs

    def sampleGiven(self, value):
        args = []
        for arg in self.arguments:
            if isinstance(arg, StarredDistribution):
                args.extend(value[arg.value])
            else:
                args.append(value[arg])
        kwargs = {name: value[arg] for name, arg in self.kwargs.items()}
        return self.method(self.object, *args, **kwargs)

    def evaluateInner(self, context):
        obj = valueInContext(self.object, context)
        arguments = tuple(valueInContext(arg, context) for arg in self.arguments)
        kwargs = {name: valueInContext(arg, context) for name, arg in self.kwargs.items()}
        return MethodDistribution(self.method, obj, arguments, kwargs)

    def __repr__(self):
        args = argsToString(self.arguments, self.kwargs)
        return f"{self.object!r}.{self.method.__name__}({args})"


def distributionMethod(method=None, *, identity=None):
    """Decorator for wrapping a method so that it can take distributions as arguments."""
    if method is None:
        return lambda method: distributionMethod(method, identity=identity)

    @unpacksDistributions
    @functools.wraps(method)
    def helper(self, *args, **kwargs):
        if identity is not None and self == identity:
            return toDistribution(args[0])
        args = tuple(toDistribution(arg) for arg in args)
        kwargs = {name: toDistribution(arg) for name, arg in kwargs.items()}
        if any(needsSampling(arg) for arg in itertools.chain(args, kwargs.values())):
            return MethodDistribution(method, self, args, kwargs)
        elif any(
            needsLazyEvaluation(arg) for arg in itertools.chain(args, kwargs.values())
        ):
            # see analogous comment in distributionFunction
            return makeDelayedFunctionCall(helper, (self,) + args, kwargs)
        else:
            return method(self, *args, **kwargs)

    return helper


class AttributeDistribution(Distribution):
    """Distribution resulting from accessing an attribute of a distribution"""

    _deterministic = True

    def __init__(self, attribute, obj, valueType=None):
        if valueType is None:
            ty = type_support.underlyingType(obj)
            valueType = self.inferType(ty, attribute)
        super().__init__(obj, valueType=valueType)
        self.attribute = attribute
        self.object = obj

    @staticmethod
    def inferType(ty, attribute):
        """Attempt to infer the type of the given attribute."""
        # If the object's type is known, see if we have an attribute type annotation.
        try:
            hints = get_type_hints(ty)
            attrTy = hints.get(attribute)
            if attrTy:
                return attrTy
        except Exception:
            pass  # couldn't get type annotations

        # Handle unions
        origin = type_support.get_type_origin(ty)
        if origin == typing.Union:
            types = []
            for option in type_support.get_type_args(ty):
                if option is type(None) and not hasattr(None, attribute):
                    # None does not have this attribute; accessing it will raise an
                    # exception, so we can ignore this case for type inference.
                    continue
                res = AttributeDistribution.inferType(option, attribute)
                types.append(res)
            return type_support.unifierOfTypes(types) if types else object

        # Check for a @property defined on the class with a return type
        if (
            ty is not object
            and (func := getattr(ty, attribute, None))
            and isinstance(func, property)
        ):
            return get_type_hints(func.fget).get("return")

        # We can't tell what the attribute type is.
        return object

    def sampleGiven(self, value):
        obj = value[self.object]
        return getattr(obj, self.attribute)

    def evaluateInner(self, context):
        obj = valueInContext(self.object, context)
        return AttributeDistribution(self.attribute, obj)

    def supportInterval(self):
        obj = self.object
        if isinstance(obj, MultiplexerDistribution):
            attrs = (getattr(opt, self.attribute) for opt in obj.options)
            return unionOfSupports(supportInterval(attr) for attr in attrs)
        return None, None

    def __call__(self, *args, **kwargs):
        vty = self.object._valueType
        retTy = None
        if vty is not object:
            func = getattr(vty, self.attribute, None)
            if func:
                if isinstance(func, property):
                    func = func.fget
                retTy = get_type_hints(func).get("return")
        return OperatorDistribution("__call__", self, args, kwargs, valueType=retTy)

    def __repr__(self):
        return f"{self.object!r}.{self.attribute}"


class OperatorDistribution(Distribution):
    """Distribution resulting from applying an operator to one or more distributions"""

    _deterministic = True

    def __init__(self, operator, obj, operands, kwoperands, valueType=None):
        operands = tuple(toDistribution(arg) for arg in operands)
        kwoperands = {key: toDistribution(kwarg) for key, kwarg in kwoperands.items()}
        dependencies = operands + tuple(kwoperands.values())
        if valueType is None:
            ty = type_support.underlyingType(obj)
            valueType = self.inferType(ty, operator, operands, kwoperands)
        super().__init__(obj, *dependencies, valueType=valueType)
        self.operator = operator
        self.object = obj
        self.operands = operands
        self.kwoperands = kwoperands
        self.symbol = allowedReversibleOperators.get(operator)
        if self.symbol:
            if operator[:3] == "__r":
                self.reverse = "__" + operator[3:]
            else:
                self.reverse = "__r" + operator[2:]
        else:
            self.reverse = None

    @staticmethod
    def inferType(ty, operator, operands, kwoperands):
        """Attempt to infer the result type of the given operator application."""
        # If the object's type is known, see if we have a return type annotation.
        origin = type_support.get_type_origin(ty)
        op = getattr(origin if origin else ty, operator, None)
        if op:
            retTy = get_type_hints(op).get("return")
            if retTy:
                return retTy

        # Handle unions
        if origin == typing.Union:
            types = []
            for option in type_support.get_type_args(ty):
                if option is type(None) and not hasattr(None, operator):
                    # None does not support this operator; using it will raise an
                    # exception, so we can ignore this case for type inference.
                    continue
                res = OperatorDistribution.inferType(
                    option, operator, operands, kwoperands
                )
                types.append(res)
            return type_support.unifierOfTypes(types) if types else object

        # The supported arithmetic operations on scalars all return scalars.
        def scalar(thing):
            ty = type_support.underlyingType(thing)
            return type_support.canCoerceType(ty, float)

        if type_support.canCoerceType(ty, float) and all(
            scalar(operand) for operand in operands
        ):
            return float

        # Handle __getitem__ for known container types.
        if operator == "__getitem__" and len(operands) == 1:
            if origin and isinstance(origin, type):  # parametrized ordinary type
                res = object
                index = operands[0]
                ity = type_support.underlyingType(index)
                isSlice = not hasattr(ity, "__index__")
                if issubclass(origin, list):
                    res = ty if isSlice else type_support.get_type_args(ty)[0]
                elif issubclass(origin, tuple):
                    args = type_support.get_type_args(ty)
                    if type_support.is_typing_generic(ty):
                        # If annotation used Tuple instead of tuple, preserve it
                        origin = typing.Tuple
                    if len(args) == 2 and args[1] is ...:
                        # homogeneous variable-length tuple
                        res = ty if isSlice else args[0]
                    elif not needsSampling(index):
                        res = args[index]
                        if isSlice:
                            res = origin[res]
                    else:
                        # can't pin down the index, so cover all possibilities
                        if any(arg is typing.Any for arg in args):
                            return origin if isSlice else object
                        res = typing.Union[args]
                        if isSlice:
                            res = origin[res, ...]
                elif issubclass(origin, (dict, typing.Mapping, collections.abc.Mapping)):
                    res = type_support.get_type_args(ty)[1]
                return object if res is typing.Any else res
            elif not origin:  # ordinary type
                if issubclass(ty, str):
                    return str

        # We can't tell what the result type is.
        return object

    def sampleGiven(self, value):
        first = value[self.object]
        rest = [value[child] for child in self.operands]
        kwargs = {key: value[child] for key, child in self.kwoperands.items()}
        op = getattr(first, self.operator)
        result = op(*rest, **kwargs)
        if result is NotImplemented and self.reverse:
            assert len(rest) == 1 and len(kwargs) == 0
            rop = getattr(rest[0], self.reverse)
            result = rop(first)
        if result is NotImplemented and self.symbol:
            raise TypeError(
                f"unsupported operand type(s) for {self.symbol}: "
                f"'{type(first).__name__}' and '{type(rest[0]).__name__}'"
            )
        return result

    def evaluateInner(self, context):
        obj = valueInContext(self.object, context)
        operands = tuple(valueInContext(arg, context) for arg in self.operands)
        kwoperands = {
            key: valueInContext(arg, context) for key, kwarg in self.kwoperands.items()
        }
        return OperatorDistribution(self.operator, obj, operands, kwoperands)

    def supportInterval(self):
        if self.operator in (
            "__add__",
            "__radd__",
            "__sub__",
            "__rsub__",
            "__mul__",
            "__rmul__",
            "__truediv__",
            "__rtruediv__",
        ):
            assert len(self.operands) == 1 and len(self.kwoperands) == 0
            l1, r1 = supportInterval(self.object)
            l2, r2 = supportInterval(self.operands[0])
            if l1 is None or l2 is None or r1 is None or r2 is None:
                return None, None
            if self.operator == "__add__" or self.operator == "__radd__":
                l = l1 + l2
                r = r1 + r2
            elif self.operator == "__sub__":
                l = l1 - r2
                r = r1 - l2
            elif self.operator == "__rsub__":
                l = l2 - r1
                r = r2 - l1
            elif self.operator in ("__mul__", "__rmul__"):
                prods = (l1 * l2, l1 * r2, r1 * l2, r1 * r2)
                l = min(*prods)
                r = max(*prods)
            elif self.operator == "__truediv__":
                if l2 > 0:
                    l = l1 / r2 if l1 >= 0 else l1 / l2
                    r = r1 / l2 if r1 >= 0 else r1 / r2
                else:
                    l, r = None, None  # TODO improve
            elif self.operator == "__rtruediv__":
                if l1 > 0:
                    l = l2 / r1 if l2 >= 0 else l2 / l1
                    r = r2 / l1 if r2 >= 0 else r2 / r1
                else:
                    l, r = None, None
            else:
                raise AssertionError(f"unexpected operator {self.operator}")
            return l, r
        elif self.operator in ("__neg__", "__abs__"):
            assert len(self.operands) == 0 and len(self.kwoperands) == 0
            l, r = supportInterval(self.object)
            if self.operator == "__neg__":
                return -r, -l
            elif self.operator == "__abs__":
                if r < 0:
                    return -r, -l
                elif l < 0:
                    return 0, max(-l, r)
                else:
                    return l, r
            else:
                raise AssertionError(f"unexpected operator {self.operator}")
        return None, None

    def __repr__(self):
        return f"{self.object!r}.{self.operator}({argsToString(self.operands)})"


# Operators which can be applied to distributions.
# Note that we deliberately do not include comparisons and __bool__,
# since Scenic does not allow control flow to depend on random variables.
allowedOperators = (
    "__neg__",
    "__pos__",
    "__abs__",
    "__round__",
    "__getitem__",
    ("__len__", int),
)
allowedReversibleOperators = {
    "__add__": "+",
    "__radd__": "+",
    "__sub__": "-",
    "__rsub__": "-",
    "__mul__": "*",
    "__rmul__": "*",
    "__truediv__": "/",
    "__rtruediv__": "/",
    "__floordiv__": "//",
    "__rfloordiv__": "//",
    "__mod__": "%",
    "__rmod__": "%",
    "__divmod__": "divmod()",
    "__rdivmod__": "divmod()",
    "__pow__": "**",
    "__rpow__": "**",
}


def makeOperatorHandler(op, ty):
    # Various special cases to simplify the expression forest by removing some
    # operations that do nothing (such as adding zero to a random number).
    if op in ("__add__", "__radd__", "__sub__"):

        def handler(self, arg):
            if (
                not isLazy(arg)
                and issubclass(self._valueType, numbers.Number)
                and arg == 0
            ):
                return self
            return OperatorDistribution(op, self, (arg,), {}, valueType=ty)

    elif op in ("__mul__", "__rmul__"):

        def handler(self, arg):
            if not isLazy(arg):
                if issubclass(self._valueType, numbers.Number) and arg == 1:
                    return self
                from scenic.core.vectors import Orientation, globalOrientation

                if issubclass(self._valueType, Orientation) and arg == globalOrientation:
                    return self
            return OperatorDistribution(op, self, (arg,), {}, valueType=ty)

    elif op in ("__truediv__", "__floordiv__", "__pow__"):

        def handler(self, arg):
            if (
                not isLazy(arg)
                and issubclass(self._valueType, numbers.Number)
                and arg == 1
            ):
                return self
            return OperatorDistribution(op, self, (arg,), {}, valueType=ty)

    else:
        # The general case.
        def handler(self, *args):
            return OperatorDistribution(op, self, args, {}, valueType=ty)

    return handler


for data in allowedOperators:
    if isinstance(data, tuple):
        op, ty = data
    else:
        op = data
        ty = None
    setattr(Distribution, op, makeOperatorHandler(op, ty))
for op in allowedReversibleOperators:
    setattr(Distribution, op, makeOperatorHandler(op, None))

import scenic.core.type_support as type_support


class MultiplexerDistribution(Distribution):
    """Distribution selecting among values based on another distribution."""

    _deterministic = True

    def __init__(self, index, options):
        self.index = index
        self.options = tuple(toDistribution(opt) for opt in options)
        assert len(self.options) > 0
        valueType = type_support.unifyingType(self.options)
        super().__init__(index, *self.options, valueType=valueType)

    def sampleGiven(self, value):
        idx = value[self.index]
        assert 0 <= idx < len(self.options), (idx, len(self.options))
        return value[self.options[idx]]

    def serializeValue(self, values, serializer):
        # We override this method to save space: we don't need to serialize all
        # of our options, only the one we're selecting.
        serializer.writeSamplable(self.index, values)
        choice = self.options[values[self.index]]
        serializer.writeSamplable(choice, values)

    def deserializeValue(self, serializer, values):
        serializer.readSamplable(self.index, values)
        choice = self.options[values[self.index]]
        serializer.readSamplable(choice, values)
        return values[choice]

    def evaluateInner(self, context):
        return type(self)(
            valueInContext(self.index, context),
            (valueInContext(opt, context) for opt in self.options),
        )

    def supportInterval(self):
        return unionOfSupports(supportInterval(opt) for opt in self.options)


## Simple distributions


class Range(Distribution):
    """Uniform distribution over a range"""

    def __init__(self, low, high):
        low = type_support.toScalar(low, f"Range endpoint {low} is not a scalar")
        high = type_support.toScalar(high, f"Range endpoint {high} is not a scalar")
        super().__init__(low, high, valueType=float)
        self.low = low
        self.high = high

    def clone(self):
        return type(self)(self.low, self.high)

    def bucket(self, buckets=None):
        if buckets is None:
            buckets = 5
        if not isinstance(buckets, int) or buckets < 1:
            raise ValueError(f"Invalid buckets for Range.bucket: {buckets}")
        if not isinstance(self.low, float) or not isinstance(self.high, float):
            raise RuntimeError(f"Cannot bucket Range with non-constant endpoints")
        endpoints = numpy.linspace(self.low, self.high, buckets + 1)
        ranges = []
        for i, left in enumerate(endpoints[:-1]):
            right = endpoints[i + 1]
            ranges.append(Range(left, right))
        return Options(ranges)

    def sampleGiven(self, value):
        return random.uniform(value[self.low], value[self.high])

    def evaluateInner(self, context):
        low = valueInContext(self.low, context)
        high = valueInContext(self.high, context)
        return Range(low, high)

    def supportInterval(self):
        return unionOfSupports((supportInterval(self.low), supportInterval(self.high)))

    def __repr__(self):
        return f"Range({self.low!r}, {self.high!r})"


class Normal(Distribution):
    """Normal distribution"""

    def __init__(self, mean, stddev):
        mean = type_support.toScalar(mean, f"Normal mean {mean} is not a scalar")
        stddev = type_support.toScalar(stddev, f"Normal stddev {stddev} is not a scalar")
        super().__init__(mean, stddev, valueType=float)
        self.mean = mean
        self.stddev = stddev

    @staticmethod
    def cdf(mean, stddev, x):
        return (1 + math.erf((x - mean) / (sqrt2 * stddev))) / 2

    @staticmethod
    def cdfinv(mean, stddev, x):
        import scipy  # slow import not often needed

        return mean + (sqrt2 * stddev * scipy.special.erfinv(2 * x - 1))

    def clone(self):
        return type(self)(self.mean, self.stddev)

    def bucket(self, buckets=None):
        if not isinstance(self.stddev, float):  # TODO relax restriction?
            raise RuntimeError(
                f"Cannot bucket Normal with non-constant standard deviation"
            )
        if buckets is None:
            buckets = 5
        if isinstance(buckets, int):
            if buckets < 1:
                raise ValueError(f"Invalid buckets for Normal.bucket: {buckets}")
            elif buckets == 1:
                endpoints = []
            elif buckets == 2:
                endpoints = [0]
            else:
                left = self.stddev * (-(buckets - 3) / 2 - 0.5)
                right = self.stddev * ((buckets - 3) / 2 + 0.5)
                endpoints = numpy.linspace(left, right, buckets - 1)
        else:
            endpoints = tuple(buckets)
            for i, v in enumerate(endpoints[:-1]):
                if v >= endpoints[i + 1]:
                    raise ValueError(
                        "Non-increasing bucket endpoints for "
                        f"Normal.bucket: {endpoints}"
                    )
        if len(endpoints) == 0:
            return Options([self.clone()])
        buckets = [(-math.inf, endpoints[0])]
        buckets.extend((v, endpoints[i + 1]) for i, v in enumerate(endpoints[:-1]))
        buckets.append((endpoints[-1], math.inf))
        pieces = []
        probs = []
        for left, right in buckets:
            pieces.append(self.mean + TruncatedNormal(0, self.stddev, left, right))
            prob = Normal.cdf(0, self.stddev, right) - Normal.cdf(0, self.stddev, left)
            probs.append(prob)
        assert math.isclose(math.fsum(probs), 1), probs
        return Options(dict(zip(pieces, probs)))

    def sampleGiven(self, value):
        return random.gauss(value[self.mean], value[self.stddev])

    def evaluateInner(self, context):
        mean = valueInContext(self.mean, context)
        stddev = valueInContext(self.stddev, context)
        return Normal(mean, stddev)

    def __repr__(self):
        return f"Normal({self.mean!r}, {self.stddev!r})"


class TruncatedNormal(Normal):
    """Truncated normal distribution."""

    def __init__(self, mean, stddev, low, high):
        if not isinstance(low, (float, int)) or not isinstance(
            high, (float, int)
        ):  # TODO relax restriction?
            raise ValueError("Endpoints of TruncatedNormal must be constant")
        if low >= high:
            raise ValueError(
                "low endpoint of TruncatedNormal must be below high endpoint"
            )
        super().__init__(mean, stddev)
        self.low = low
        self.high = high

    def clone(self):
        return type(self)(self.mean, self.stddev, self.low, self.high)

    def bucket(self, buckets=None):
        if not isinstance(self.stddev, float):  # TODO relax restriction?
            raise RuntimeError(
                "Cannot bucket TruncatedNormal with " "non-constant standard deviation"
            )
        if buckets is None:
            buckets = 5
        if isinstance(buckets, int):
            if buckets < 1:
                raise ValueError(f"Invalid buckets for TruncatedNormal.bucket: {buckets}")
            endpoints = numpy.linspace(self.low, self.high, buckets + 1)
        else:
            endpoints = tuple(buckets)
            if len(endpoints) < 2:
                raise ValueError(
                    "Too few bucket endpoints for " f"TruncatedNormal.bucket: {endpoints}"
                )
            if endpoints[0] != self.low or endpoints[-1] != self.high:
                raise ValueError(
                    f"TruncatedNormal.bucket endpoints {endpoints} " "do not match domain"
                )
            for i, v in enumerate(endpoints[:-1]):
                if v >= endpoints[i + 1]:
                    raise ValueError(
                        "Non-increasing bucket endpoints for "
                        f"TruncatedNormal.bucket: {endpoints}"
                    )
        pieces, probs = [], []
        for i, left in enumerate(endpoints[:-1]):
            right = endpoints[i + 1]
            pieces.append(TruncatedNormal(self.mean, self.stddev, left, right))
            prob = Normal.cdf(self.mean, self.stddev, right) - Normal.cdf(
                self.mean, self.stddev, left
            )
            probs.append(prob)
        return Options(dict(zip(pieces, probs)))

    def sampleGiven(self, value):
        # TODO switch to method less prone to underflow?
        mean, stddev = value[self.mean], value[self.stddev]
        alpha = (self.low - mean) / stddev
        beta = (self.high - mean) / stddev
        alpha_cdf = Normal.cdf(0, 1, alpha)
        beta_cdf = Normal.cdf(0, 1, beta)
        if beta_cdf - alpha_cdf < 1e-15:
            warnings.warn("low precision when sampling TruncatedNormal")
        unif = random.random()
        p = alpha_cdf + unif * (beta_cdf - alpha_cdf)
        return mean + (stddev * Normal.cdfinv(0, 1, p))

    def evaluateInner(self, context):
        mean = valueInContext(self.mean, context)
        stddev = valueInContext(self.stddev, context)
        return TruncatedNormal(mean, stddev, self.low, self.high)

    def supportInterval(self):
        return self.low, self.high

    def __repr__(self):
        return f"TruncatedNormal({self.mean!r}, {self.stddev!r}, {self.low!r}, {self.high!r})"


class DiscreteRange(Distribution):
    """Distribution over a range of integers."""

    def __init__(self, low, high, weights=None, emptyMessage="empty DiscreteRange"):
        if weights:
            if not isinstance(low, int):
                raise TypeError("weighted DiscreteRange left endpoint must be an integer")
            if not isinstance(high, int):
                raise TypeError(
                    "weighted DiscreteRange right endpoint must be an integer"
                )
            if not low <= high:
                raise ValueError(
                    f"DiscreteRange lower bound {low} is above upper bound {high}"
                )
            self.low, self.high = low, high
            weights = tuple(weights)
            if len(weights) != high - low + 1:
                raise ValueError("weighted DiscreteRange with wrong number of weights")
            self.weights = weights
            self.cumulativeWeights = tuple(itertools.accumulate(weights))
            self.options = tuple(range(low, high + 1))
        else:
            self.low = type_support.toScalar(
                low, "DiscreteRange left endpoint not a scalar"
            )
            self.high = type_support.toScalar(
                high, "DiscreteRange right endpoint not a scalar"
            )
            self.weights = None
        self.emptyMessage = emptyMessage
        super().__init__(self.low, self.high, valueType=int)

    def clone(self):
        return type(self)(self.low, self.high, self.weights, self.emptyMessage)

    def bucket(self, buckets=None):
        return self.clone()  # already bucketed

    def sampleGiven(self, value):
        if self.weights:
            return random.choices(self.options, cum_weights=self.cumulativeWeights)[0]
        left, right = math.ceil(value[self.low]), math.floor(value[self.high])
        if right < left:
            raise RejectionException(self.emptyMessage)
        return random.randint(left, right)

    def supportInterval(self):
        ll, lh = supportInterval(self.low)
        hl, hh = supportInterval(self.high)
        return ll, hh

    def __repr__(self):
        weights = self.weights
        if all(weight == weights[0] for weight in weights):
            return f"DiscreteRange({self.low!r}, {self.high!r})"
        else:
            return f"DiscreteRange({self.low!r}, {self.high!r}, {self.weights})"


class Options(MultiplexerDistribution):
    """Distribution over a finite list of options.

    Specified by a dict giving probabilities; otherwise uniform over a given iterable.
    """

    def __init__(self, opts):
        if isinstance(opts, dict):
            options, weights = [], []
            for opt, prob in opts.items():
                if not isinstance(prob, (float, int)):
                    raise TypeError(
                        f"discrete distribution weight {prob}" " is not a constant number"
                    )
                if prob < 0:
                    raise ValueError(f"discrete distribution weight {prob} is negative")
                if prob == 0:
                    continue
                options.append(opt)
                weights.append(prob)
            self.optWeights = dict(zip(options, weights))
        else:
            weights = None
            options = tuple(opts)
            self.optWeights = None
        if len(options) == 0:
            raise RejectionException(
                "tried to make discrete distribution over empty domain!"
            )

        index = self.makeSelector(len(options) - 1, weights)
        super().__init__(index, options)

    @staticmethod
    def makeSelector(n, weights):
        return DiscreteRange(0, n, weights)

    def clone(self):
        return type(self)(self.optWeights if self.optWeights else self.options)

    def bucket(self, buckets=None):
        return self.clone()  # already bucketed

    def evaluateInner(self, context):
        if self.optWeights is None:
            return type(self)(valueInContext(opt, context) for opt in self.options)
        else:
            return type(self)(
                {valueInContext(opt, context): wt for opt, wt in self.optWeights.items()}
            )

    def __repr__(self):
        if self.optWeights is not None:
            return f"{type(self).__name__}({self.optWeights!r})"
        else:
            args = ", ".join(repr(opt) for opt in self.options)
            return f"{type(self).__name__}({args})"


@unpacksDistributions
def Uniform(*opts):
    """Uniform distribution over a finite list of options.

    Implemented as an instance of :obj:`Options` when the set of options is known
    statically, and an instance of `UniformDistribution` otherwise.
    """
    if any(isinstance(opt, StarredDistribution) for opt in opts):
        return UniformDistribution(opts)
    else:
        return Options(opts)


class UniformDistribution(Distribution):
    """Uniform distribution over a variable number of options.

    See :obj:`Options` for the more common uniform distribution over a fixed number
    of options. This class is for the special case where iterable unpacking is
    applied to a distribution, so that the number of options is unknown at
    compile time.
    """

    _deterministic = True

    def __init__(self, opts):
        self.options = opts
        valueType = type_support.unifyingType(self.options)
        length = 0
        for opt in opts:
            if isinstance(opt, StarredDistribution):
                length += opt.__len__()
            else:
                length += 1
        self.selector = DiscreteRange(
            0, length - 1, emptyMessage="uniform distribution over empty domain"
        )
        super().__init__(*self.options, self.selector, valueType=valueType)

    def clone(self):
        return type(self)(self.options)

    def sampleGiven(self, value):
        index = value[self.selector]
        opts = []
        for opt in self.options:
            if isinstance(opt, StarredDistribution):
                opts.extend(value[opt])
            else:
                opts.append(value[opt])
        assert 0 <= index < len(opts)
        return opts[index]

    def evaluateInner(self, context):
        opts = tuple(valueInContext(opt, context) for opt in self.options)
        return UniformDistribution(opts)

    def __repr__(self):
        return f"UniformDistribution({self.options!r})"



"""Python supporting code for the GTA model."""

from dataclasses import dataclass
import math
import random
import time

import numpy
import scipy.spatial

try:
    import PIL
except ModuleNotFoundError as e:
    raise RuntimeError(
        "GTA scenarios require the PIL module;" ' try "pip install pillow"'
    ) from e

try:
    import cv2
except ModuleNotFoundError as e:
    raise RuntimeError(
        "GTA scenarios require the cv2 module;" ' try "pip install opencv-python"'
    ) from e

from scenic.core.distributions import (
    Distribution,
    Normal,
    Options,
    Range,
    Samplable,
    distributionMethod,
    toDistribution,
)
from scenic.core.geometry import *
from scenic.core.lazy_eval import valueInContext
from scenic.core.regions import GridRegion, PointSetRegion
import scenic.core.utils as utils
from scenic.core.vectors import VectorField
from scenic.core.workspaces import Workspace
import scenic.simulators.gta.center_detection as center_detection
import scenic.simulators.gta.img_modf as img_modf
import scenic.simulators.gta.messages as messages
from scenic.simulators.utils.colors import Color
from scenic.syntax.veneer import verbosePrint

### Abstract GTA interface


class GTA:
    @staticmethod
    def Config(scene):
        ego = scene.egoObject
        vehicles = [GTA.Vehicle(car) for car in scene.objects if car is not ego]
        cameraLoc = GTA.langToGTACoords(ego.position)
        cameraHeading = GTA.langToGTAHeading(ego.heading)

        params = dict(scene.params)
        time = int(round(params.pop("time"))) % 1440
        hour, minute = divmod(time, 60)
        assert hour < 24, scene.params["time"]
        weather = params.pop("weather")
        for param in params:
            print(f'WARNING: unused scene parameter "{param}"')

        return messages.Formal_Config(
            cameraLoc, [hour, minute], weather, vehicles, cameraHeading
        )

    @staticmethod
    def Vehicle(car):
        loc3 = GTA.langToGTACoords(car.position)
        heading = GTA.langToGTAHeading(car.heading)
        scol = list(Color.realToByte(car.color))
        return messages.Vehicle(car.model.name, scol, loc3, heading)

    @staticmethod
    def langToGTACoords(point):
        x, y, _ = point
        return [x, y, 60]

    @staticmethod
    def langToGTAHeading(heading):
        h = math.degrees(heading)
        return (h + 360) % 360


### Map


class Map:
    """Represents roads and obstacles in GTA, extracted from a map image.

    This code handles images from the `GTA V Interactive Map <https://gta-5-map.com/>`_,
    rendered with the "Road" setting.

    Args:
        imagePath (str): path to image file
        Ax (float): width of one pixel in GTA coordinates
        Ay (float): height of one pixel in GTA coordinates
        Bx (float): GTA X-coordinate of bottom-left corner of image
        By (float): GTA Y-coordinate of bottom-left corner of image
    """

    def __init__(self, imagePath, Ax, Ay, Bx, By):
        self.Ax, self.Ay = Ax, Ay
        self.Bx, self.By = Bx, By
        if imagePath != None:
            startTime = time.time()
            # open image
            image = PIL.Image.open(imagePath)
            self.sizeX, self.sizeY = image.size
            # create version of image for display
            de = img_modf.get_edges(image).convert("RGB")
            self.displayImage = cv2.cvtColor(numpy.array(de), cv2.COLOR_RGB2BGR)
            # detect edges of roads
            ed = center_detection.compute_midpoints(img_data=image, kernelsize=5)
            edges, tangents = [], []
            for (y, x), tangent in ed.items():
                edges.append(self.gridToScenicCoords((x, y)))
                tangents.append(tangent)
            self.edges = numpy.array(edges)
            self.edgeTangents = numpy.array(tangents)
            # build k-D tree
            self.edgeTree = scipy.spatial.KDTree(self.edges)
            # identify points on roads
            self.roadArray = numpy.array(
                img_modf.convert_black_white(img_data=image).convert("L"), dtype=int
            )
            totalTime = time.time() - startTime
            verbosePrint(f"Created GTA map from image in {totalTime:.2f} seconds.")

    @staticmethod
    def fromFile(path):
        startTime = time.time()
        with numpy.load(path) as data:
            Ax, Ay, Bx, By, sizeX, sizeY = data["misc"]
            m = Map(None, Ax, Ay, Bx, By)
            m.sizeX, m.sizeY = sizeX, sizeY
            m.displayImage = data["displayImage"]

            m.edges = data["edges"]
            m.edgeTangents = data["tangents"]
            m.edgeTree = scipy.spatial.KDTree(m.edges)  # rebuild k-D tree

            m.roadArray = data["roadArray"]
            totalTime = time.time() - startTime
            verbosePrint(f"Loaded GTA map in {totalTime:.2f} seconds.")
            return m

    def dumpToFile(self, path):
        misc = numpy.array((self.Ax, self.Ay, self.Bx, self.By, self.sizeX, self.sizeY))

        numpy.savez_compressed(
            path,
            misc=misc,
            displayImage=self.displayImage,
            edges=self.edges,
            tangents=self.edgeTangents,
            roadArray=self.roadArray,
        )

    @property
    @utils.cached
    def roadDirection(self):
        return VectorField("roadDirection", self.roadHeadingAt)

    @property
    @utils.cached
    def roadRegion(self):
        return GridRegion(
            "road",
            self.roadArray,
            self.Ax,
            self.Ay,
            self.Bx,
            self.By,
            orientation=self.roadDirection,
        )

    @property
    @utils.cached
    def curbRegion(self):
        return PointSetRegion("curb", self.edges, orientation=self.roadDirection)

    def gridToScenicCoords(self, point):
        x, y = point[0], point[1]
        return ((self.Ax * x) + self.Bx, (self.Ay * y) + self.By)

    def gridToScenicHeading(self, heading):
        return heading - (math.pi / 2)

    def scenicToGridCoords(self, point):
        x, y = point[0], point[1]
        return ((x - self.Bx) / self.Ax, (y - self.By) / self.Ay)

    def scenicToGridHeading(self, heading):
        return heading + (math.pi / 2)

    @distributionMethod
    def roadHeadingAt(self, point):
        # find closest edge
        point2d = (point[0], point[1])
        distance, location = self.edgeTree.query(point2d)
        # get direction of edge
        return self.gridToScenicHeading(self.edgeTangents[location])

    def show(self, plt):
        plt.imshow(self.displayImage)


class MapWorkspace(Workspace):
    """Workspace whose rendering is handled by a Map"""

    def __init__(self, mappy, region):
        super().__init__(region)
        self.map = mappy

    def scenicToSchematicCoords(self, coords):
        return self.map.scenicToGridCoords(coords)

    def show2D(self, plt):
        return self.map.show(plt)

    @property
    def minimumZoomSize(self):
        return 40 / min(abs(self.map.Ax), abs(self.map.Ay))


### Car models and colors


@dataclass(frozen=True)
class CarModel:
    """A model of car in GTA.

    Attributes:
        name (str): name of model in GTA
        width (float): width of this model of car
        length (float): length of this model of car
        viewAngle (float): view angle in radians (default is 90 degrees)

    Class Attributes:
        models: dict mapping model names to the corresponding `CarModel`
    """

    name: str
    width: float
    length: float
    viewAngle: float = math.radians(90)

    @classmethod
    def uniformModel(self):
        return Options(self.modelProbs.keys())

    @classmethod
    def egoModel(self):
        return self.models["BLISTA"]

    @classmethod
    def defaultModel(self):
        return Options(self.modelProbs)

    def __str__(self):
        return f"<CarModel {self.name}>"


CarModel.modelProbs = {
    CarModel("BLISTA", 1.75871, 4.10139): 1,
    CarModel("BUS", 2.9007, 13.202): 0,
    CarModel("NINEF", 2.07699, 4.50658): 1,
    CarModel("ASEA", 1.83066, 4.45861): 1,
    CarModel("BALLER", 2.10791, 5.10333): 1,
    CarModel("BISON", 2.29372, 5.4827): 1,
    CarModel("BUFFALO", 2.04265, 5.07782): 1,
    CarModel("BOBCATXL", 2.37944, 5.78222): 1,
    CarModel("DOMINATOR", 1.9353, 4.9355): 1,
    CarModel("GRANGER", 3.02698, 5.94577): 1,
    CarModel("JACKAL", 2.00041, 4.91436): 1,
    CarModel("ORACLE", 2.07787, 5.12544): 1,
    CarModel("PATRIOT", 2.26679, 5.13695): 1,
    CarModel("PRANGER", 3.02698, 5.94577): 1,
}
CarModel.models = {model.name: model for model in CarModel.modelProbs}

CarColor = Color  # for backwards compatibility



"""Newtonian simulator implementation."""

from cmath import atan, pi, tan
import math
from math import copysign, degrees, radians, sin
import os
import pathlib
import time

from PIL import Image
import numpy as np

import scenic.core.errors as errors  # isort: skip

if errors.verbosityLevel == 0:  # suppress pygame advertisement at zero verbosity
    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
import pygame
import shapely

from scenic.core.geometry import allChains, findMinMax
from scenic.core.regions import toPolygon
from scenic.core.simulators import SimulationCreationError
from scenic.core.vectors import Orientation, Vector
from scenic.domains.driving.controllers import (
    PIDLateralController,
    PIDLongitudinalController,
)
from scenic.domains.driving.roads import Network
from scenic.domains.driving.simulators import DrivingSimulation, DrivingSimulator
from scenic.syntax.veneer import verbosePrint

current_dir = pathlib.Path(__file__).parent.absolute()

WIDTH = 1280
HEIGHT = 800
MAX_ACCELERATION = 5.6  # in m/s2, seems to be a pretty reasonable value
MAX_BRAKING = 4.6

ROAD_COLOR = (0, 0, 0)
ROAD_WIDTH = 2
LANE_COLOR = (96, 96, 96)
CENTERLINE_COLOR = (224, 224, 224)
SIDEWALK_COLOR = (0, 128, 255)
SHOULDER_COLOR = (96, 96, 96)


class NewtonianSimulator(DrivingSimulator):
    """Implementation of `Simulator` for the Newtonian simulator.

    Args:
        network (Network): road network to display in the background, if any.
        render (bool): whether to render the simulation in a window.

    .. versionchanged:: 3.0

        The **timestep** argument is removed: it can be specified when calling
        `simulate` instead. The default timestep for the Newtonian simulator
        when not otherwise specified is still 0.1 seconds.
    """

    def __init__(self, network=None, render=False, export_gif=False):
        super().__init__()
        self.export_gif = export_gif
        self.render = render
        self.network = network

    def createSimulation(self, scene, **kwargs):
        simulation = NewtonianSimulation(
            scene, self.network, self.render, self.export_gif, **kwargs
        )
        if self.export_gif and self.render:
            simulation.generate_gif("simulation.gif")
        return simulation


class NewtonianSimulation(DrivingSimulation):
    """Implementation of `Simulation` for the Newtonian simulator."""

    def __init__(self, scene, network, render, export_gif, timestep, **kwargs):
        self.export_gif = export_gif
        self.render = render
        self.network = network
        self.frames = []

        if timestep is None:
            timestep = 0.1

        super().__init__(scene, timestep=timestep, **kwargs)

    def setup(self):
        super().setup()

        if self.render:
            # determine window size
            min_x, max_x = findMinMax(obj.x for obj in self.objects)
            min_y, max_y = findMinMax(obj.y for obj in self.objects)

            pygame.init()
            pygame.font.init()
            self.screen = pygame.display.set_mode(
                (WIDTH, HEIGHT), pygame.HWSURFACE | pygame.DOUBLEBUF
            )
            self.screen.fill((255, 255, 255))
            x, y, _ = self.objects[0].position
            self.min_x, self.max_x = min_x - 50, max_x + 50
            self.min_y, self.max_y = min_y - 50, max_y + 50
            self.size_x = self.max_x - self.min_x
            self.size_y = self.max_y - self.min_y
            self.screen_poly = shapely.geometry.Polygon(
                (
                    (self.min_x, self.min_y),
                    (self.max_x, self.min_y),
                    (self.max_x, self.max_y),
                    (self.min_x, self.max_y),
                )
            )

            img_path = os.path.join(current_dir, "car.png")
            self.car = pygame.image.load(img_path)
            self.car_width = int(3.5 * WIDTH / self.size_x)
            self.car_height = self.car_width
            self.car = pygame.transform.scale(self.car, (self.car_width, self.car_height))
            self.parse_network()
            self.draw_objects()

    def parse_network(self):
        self.network_polygons = []
        if not self.network:
            return

        def addRegion(region, color, width=1):
            poly = toPolygon(region)
            if not poly or not self.screen_poly.intersects(poly):
                return
            for chain in allChains(poly):
                coords = tuple(self.scenicToScreenVal(pt) for pt in chain.coords)
                self.network_polygons.append((coords, color, width))

        addRegion(self.network.walkableRegion, SIDEWALK_COLOR)
        addRegion(self.network.shoulderRegion, SHOULDER_COLOR)
        for road in self.network.roads:  # loop over main roads
            for lane in road.lanes:
                addRegion(lane.leftEdge, LANE_COLOR)
                addRegion(lane.rightEdge, LANE_COLOR)
            addRegion(road, ROAD_COLOR, ROAD_WIDTH)
        for lane in self.network.lanes:  # loop over all lanes, even in intersections
            addRegion(lane.centerline, CENTERLINE_COLOR)
        addRegion(self.network.intersectionRegion, ROAD_COLOR)

    def scenicToScreenVal(self, pos):
        x, y = pos[:2]
        x_prop = (x - self.min_x) / self.size_x
        y_prop = (y - self.min_y) / self.size_y
        return int(x_prop * WIDTH), HEIGHT - 1 - int(y_prop * HEIGHT)

    def createObjectInSimulator(self, obj):
        # Set actor's initial speed
        obj.speed = math.hypot(*obj.velocity)

        if hasattr(obj, "elevation"):
            obj.elevation = 0.0

    def isOnScreen(self, x, y):
        return self.min_x <= x <= self.max_x and self.min_y <= y <= self.max_y

    def step(self):
        for obj in self.objects:
            current_speed = obj.velocity.norm()
            if hasattr(obj, "hand_brake"):
                forward = obj.velocity.dot(Vector(0, 1).rotatedBy(obj.heading)) >= 0
                signed_speed = current_speed if forward else -current_speed
                if obj.hand_brake or obj.brake > 0:
                    braking = MAX_BRAKING * max(obj.hand_brake, obj.brake)
                    acceleration = braking * self.timestep
                    if acceleration >= current_speed:
                        signed_speed = 0
                    elif forward:
                        signed_speed -= acceleration
                    else:
                        signed_speed += acceleration
                else:
                    acceleration = obj.throttle * MAX_ACCELERATION
                    if obj.reverse:
                        acceleration *= -1
                    signed_speed += acceleration * self.timestep

                obj.velocity = Vector(0, signed_speed).rotatedBy(obj.heading)
                if obj.steer:
                    turning_radius = obj.length / sin(obj.steer * math.pi / 2)
                    obj.angularSpeed = -signed_speed / turning_radius
                else:
                    obj.angularSpeed = 0
                obj.speed = abs(signed_speed)
            else:
                obj.speed = current_speed
            obj.position += obj.velocity * self.timestep
            obj.heading += obj.angularSpeed * self.timestep

        if self.render:
            self.draw_objects()
            pygame.event.pump()

    def draw_objects(self):
        self.screen.fill((255, 255, 255))
        for screenPoints, color, width in self.network_polygons:
            pygame.draw.lines(self.screen, color, False, screenPoints, width=width)

        for i, obj in enumerate(self.objects):
            color = (255, 0, 0) if i == 0 else (0, 0, 255)
            h, w = obj.length, obj.width
            pos_vec = Vector(-1.75, 1.75)
            neg_vec = Vector(w / 2, h / 2)
            heading_vec = Vector(0, 10).rotatedBy(obj.heading)
            dx, dy = int(heading_vec.x), -int(heading_vec.y)
            x, y = self.scenicToScreenVal(obj.position)
            rect_x, rect_y = self.scenicToScreenVal(obj.position + pos_vec)
            if hasattr(obj, "isCar") and obj.isCar:
                self.rotated_car = pygame.transform.rotate(
                    self.car, math.degrees(obj.heading)
                )
                self.screen.blit(self.rotated_car, (rect_x, rect_y))
            else:
                corners = [self.scenicToScreenVal(corner) for corner in obj._corners2D]
                pygame.draw.polygon(self.screen, color, corners)

        pygame.display.update()

        if self.export_gif:
            frame = pygame.surfarray.array3d(self.screen)
            frame = np.transpose(frame, (1, 0, 2))
            self.frames.append(frame)

        time.sleep(self.timestep)

    def generate_gif(self, filename="simulation.gif"):
        imgs = [Image.fromarray(frame) for frame in self.frames]
        imgs[0].save(filename, save_all=True, append_images=imgs[1:], duration=50, loop=0)

    def getProperties(self, obj, properties):
        yaw, _, _ = obj.parentOrientation.globalToLocalAngles(obj.heading, 0, 0)

        values = dict(
            position=obj.position,
            yaw=yaw,
            pitch=0,
            roll=0,
            velocity=obj.velocity,
            speed=obj.speed,
            angularSpeed=obj.angularSpeed,
            angularVelocity=obj.angularVelocity,
        )
        if "elevation" in properties:
            values["elevation"] = obj.elevation
        return values

    def destroy(self):
        if self.render:
            pygame.quit()

    def getLaneFollowingControllers(self, agent):
        dt = self.timestep
        if agent.isCar:
            lon_controller = PIDLongitudinalController(K_P=0.5, K_D=0.1, K_I=0.7, dt=dt)
            lat_controller = PIDLateralController(K_P=0.1, K_D=0.1, K_I=0.02, dt=dt)
        else:
            lon_controller = PIDLongitudinalController(
                K_P=0.25, K_D=0.025, K_I=0.0, dt=dt
            )
            lat_controller = PIDLateralController(K_P=0.2, K_D=0.1, K_I=0.0, dt=dt)
        return lon_controller, lat_controller

    def getTurningControllers(self, agent):
        dt = self.timestep
        if agent.isCar:
            lon_controller = PIDLongitudinalController(K_P=0.5, K_D=0.1, K_I=0.7, dt=dt)
            lat_controller = PIDLateralController(K_P=0.2, K_D=0.2, K_I=0.2, dt=dt)
        else:
            lon_controller = PIDLongitudinalController(
                K_P=0.25, K_D=0.025, K_I=0.0, dt=dt
            )
            lat_controller = PIDLateralController(K_P=0.4, K_D=0.1, K_I=0.0, dt=dt)
        return lon_controller, lat_controller

    def getLaneChangingControllers(self, agent):
        dt = self.timestep
        if agent.isCar:
            lon_controller = PIDLongitudinalController(K_P=0.5, K_D=0.1, K_I=0.7, dt=dt)
            lat_controller = PIDLateralController(K_P=0.2, K_D=0.2, K_I=0.02, dt=dt)
        else:
            lon_controller = PIDLongitudinalController(
                K_P=0.25, K_D=0.025, K_I=0.0, dt=dt
            )
            lat_controller = PIDLateralController(K_P=0.1, K_D=0.3, K_I=0.0, dt=dt)
        return lon_controller, lat_controller



"""Actions taken by dynamic agents."""

import abc


class Action(abc.ABC):
    """An :term:`action` which can be taken by an agent for one step of a simulation."""

    def canBeTakenBy(self, agent):
        """Whether this action is allowed to be taken by the given agent.

        The default implementation always returns True.
        """
        return True

    @abc.abstractmethod
    def applyTo(self, agent, simulation):
        """Apply this action to the given agent in the given simulation.

        This method should call simulator APIs so that the agent will take this action
        during the next simulated time step. Depending on the simulator API, it may be
        necessary to batch each agent's actions into a single update: in that case you
        can have this method set some state on the agent, then apply the actual update
        in an overridden implementation of `Simulation.executeActions`. For examples,
        see the CARLA interface: `scenic.simulators.carla.actions` has some CARLA-specific
        actions which directly call CARLA APIs, while the generic steering and braking
        actions from `scenic.domains.driving.actions` are implemented using the batching
        approach (see for example the ``setThrottle`` method of the class
        `scenic.simulators.carla.model.Vehicle`, which sets state later read by
        ``CarlaSimulation.executeActions`` in `scenic.simulators.carla.simulator`).
        """
        raise NotImplementedError


class _EndSimulationAction(Action):
    """Special action indicating it is time to end the simulation.

    Only for internal use.
    """

    def __init__(self, line):
        self.line = line

    def __str__(self):
        return f'"terminate simulation" executed on line {self.line}'

    def applyTo(self, agent, simulation):
        assert False


class _EndScenarioAction(Action):
    """Special action indicating it is time to end the current scenario.

    Only for internal use.
    """

    def __init__(self, scenario, line):
        self.scenario = scenario
        self.line = line

    def __str__(self):
        return f'"terminate" executed on line {self.line}'

    def applyTo(self, agent, simulation):
        assert False



"""Interface between Scenic and simulators.

This module defines the core classes `Simulator` and `Simulation` which
orchestrate dynamic simulations. Each simulator interface defines subclasses
of these classes for their particular simulator.

Ordinary Scenic users only need to know about the top-level simulation API
`Simulator.simulate` and the attributes of the `Simulation` class (in particular
the ``result`` attribute, which captures information about the result of the
simulation as a `SimulationResult` object).
"""

import abc
from collections import defaultdict
import enum
import math
import numbers
import time
import types

from scenic.core.distributions import RejectionException
from scenic.core.dynamics import GuardViolation, RejectSimulationException
from scenic.core.dynamics.actions import Action, _EndScenarioAction, _EndSimulationAction
import scenic.core.errors as errors
from scenic.core.errors import InvalidScenarioError, optionallyDebugRejection
from scenic.core.object_types import (
    Object2D,
    disableDynamicProxyFor,
    enableDynamicProxyFor,
    setDynamicProxyFor,
)
from scenic.core.requirements import RequirementType
from scenic.core.serialization import Serializer
from scenic.core.vectors import Vector


class SimulatorInterfaceWarning(UserWarning):
    """Warning indicating an issue with the interface to an external simulator."""

    pass


class SimulationCreationError(Exception):
    """Exception indicating a simulation could not be run from the given scene.

    Can also be issued during a simulation if dynamic object creation fails.
    """

    pass


class DivergenceError(Exception):
    """Exception indicating simulation replay failed due to simulator nondeterminism."""

    pass


class Simulator(abc.ABC):
    """A simulator which can execute dynamic simulations from Scenic scenes.

    Simulator interfaces which support dynamic simulations should implement a
    subclass of `Simulator`. An instance of the class represents a connection to
    the simulator suitable for running multiple simulations (not necessarily of
    the same Scenic program). For an example of how to implement this class,
    and its counterpart `Simulation` for individual simulations, see
    :mod:`scenic.simulators.webots.simulator`.

    Users who create an instance of `Simulator` should call its `destroy` method
    when they are finished running simulations to allow the interface to do any
    necessary cleanup.
    """

    def __init__(self):
        self._destroyed = False

    def simulate(
        self,
        scene,
        maxSteps=None,
        maxIterations=1,
        *,
        timestep=None,
        verbosity=None,
        raiseGuardViolations=False,
        replay=None,
        enableReplay=True,
        enableDivergenceCheck=False,
        divergenceTolerance=0,
        continueAfterDivergence=False,
        allowPickle=False,
    ):
        """Run a simulation for a given scene.

        For details on how simulations are run, see `dynamic scenario semantics`.

        Args:
            scene (Scene): Scene from which to start the simulation (sampled using
                `Scenario.generate`).
            maxSteps (int): Maximum number of time steps for the simulation, or `None` to
                not impose a time bound.
            maxIterations (int): Maximum number of rejection sampling iterations.
            timestep (float): Length of a time step in seconds, or `None` to use a
                default provided by the simulator interface. Some interfaces may not
                allow arbitrary time step lengths or may require the timestep to be set
                when creating the `Simulator` and not customized per-simulation.
            verbosity (int): If not `None`, override Scenic's global verbosity level
                (from the :option:`--verbosity` option or `scenic.setDebuggingOptions`).
            raiseGuardViolations (bool): Whether violations of preconditions/invariants
                of scenarios/behaviors should cause this method to raise an exception,
                instead of only rejecting the simulation (the default behavior).
            replay (bytes): If not `None`, must be replay data output by `Simulation.getReplay`:
                we will then replay the saved simulation rather than randomly generating
                one as usual. If **maxSteps** is larger than that of the original
                simulation, then once the replay is exhausted the simulation will continue
                to run in the usual randomized manner.
            enableReplay (bool): Whether to save data from the simulation so that it can
                be serialized for later replay using `Scenario.simulationToBytes` or
                `Simulation.getReplay`. Enabled by default as the overhead is generally low.
            enableDivergenceCheck (bool): Whether to save the values of every
                :term:`dynamic property` at each time step, so that when the simulation is
                replayed, nondeterminism in the simulator (or replaying the simulation in
                the wrong simulator) can be detected. Disabled by default as this option
                greatly increases the size of replay objects (~100 bytes per object per step).
            divergenceTolerance (float): Amount by which a dynamic property can deviate in
                a replay from its original value before we consider the replay to have
                diverged. The default value is zero: no deviation is allowed. If finer
                control over divergences is required, see `Simulation.valuesHaveDiverged`.
            continueAfterDivergence (bool): Whether to continue simulating after a
                divergence is detected instead of raising a `DivergenceError`. If this is
                true, then a divergence ends the replaying of the saved scenario but the
                simulation will continue in the usual randomized manner (i.e., it is as
                if the replay data ran out at the moment of the divergence).
            allowPickle (bool): Whether to use `pickle` to (de)serialize custom object
                types. See `sceneFromBytes` for a discussion of when this may be needed
                (rarely) and its security implications.

        Returns:
            A `Simulation` object representing the completed simulation, or `None` if no
            simulation satisfying the requirements could be found within
            **maxIterations** iterations.

        Raises:
            SimulationCreationError: if an error occurred while trying to run a
                simulation (e.g. some assumption made by the simulator was violated, like
                trying to create an object inside another).
            GuardViolation: if **raiseGuardViolations** is true and a precondition or
                invariant was violated during the simulation.
            DivergenceError: if replaying a simulation (via the **replay** option) and
                the replay has diverged from the original; requires the original simulation
                to have been run with **enableDivergenceCheck**.
            SerializationError: if writing or reading replay data fails. This could happen
                if your scenario uses an unusual custom distribution (see `sceneToBytes`)
                or if the replayed scenario has diverged without divergence-checking
                enabled.

        .. versionchanged:: 3.0

            **maxIterations** is now 1 by default.

        .. versionadded:: 3.0

            The **timestep** argument.
        """

        if self._destroyed:
            raise RuntimeError(
                "simulator cannot run additional simulations "
                "(the destroy() method has already been called)"
            )
        if verbosity is None:
            verbosity = errors.verbosityLevel

        # Repeatedly run simulations until we find one satisfying the requirements
        iterations = 0
        simulation = None
        while not simulation and (maxIterations is None or iterations < maxIterations):
            iterations += 1
            simulation = self._runSingleSimulation(
                scene,
                maxSteps,
                name=iterations,
                verbosity=verbosity,
                timestep=timestep,
                raiseGuardViolations=raiseGuardViolations,
                replay=replay,
                enableReplay=enableReplay,
                enableDivergenceCheck=enableDivergenceCheck,
                divergenceTolerance=divergenceTolerance,
                continueAfterDivergence=continueAfterDivergence,
                allowPickle=allowPickle,
            )
        return simulation

    def replay(self, scene, replay, **kwargs):
        """Replay a simulation.

        This convenience method simply calls `simulate` (and so takes all the same
        arguments), but makes the **replay** argument positional so you can write
        ``simulator.replay(scene, replay)`` instead of
        ``simulator.simulate(scene, replay=replay)``.
        """
        return self.simulate(scene, replay=replay, **kwargs)

    def _runSingleSimulation(
        self, scene, maxSteps, *, name, verbosity, raiseGuardViolations=False, **kwargs
    ):
        if verbosity >= 2:
            print(f"  Starting simulation {name}...")
        try:
            simulation = self.createSimulation(
                scene,
                maxSteps=maxSteps,
                name=name,
                verbosity=verbosity,
                **kwargs,
            )
        except (RejectSimulationException, RejectionException, GuardViolation) as e:
            if verbosity >= 2:
                print(
                    f"  Rejected simulation {name} at time step "
                    f"{e.simulation.currentTime} because: {e}"
                )
            if raiseGuardViolations and isinstance(e, GuardViolation):
                raise
            else:
                optionallyDebugRejection(e)
                return None

        # Completed the simulation without violating a requirement
        if not isinstance(simulation, Simulation):
            raise TypeError(f"simulator returned non-Simulation {simulation}")
        if verbosity >= 2:
            print(
                f"  Simulation {name} ended successfully at time step "
                f"{simulation.currentTime} because: {simulation.result.terminationReason}"
            )
        return simulation

    @abc.abstractmethod
    def createSimulation(self, scene, **kwargs):
        """Create a `Simulation` from a Scenic scene.

        This should be overridden by subclasses to return instances of their own
        specialized subclass of `Simulation`. The given **scene** and **kwargs**
        (together making up all the arguments passed to `simulate` except for
        **maxIterations**) should be passed through to the initializer of that
        instance.

        .. versionchanged:: 3.0

            This method is now called with all the arguments to `simulate` except for
            **maxIterations**; these should be passed through as described above.
        """
        return Simulation(scene, **kwargs)

    def destroy(self):
        """Clean up as needed when shutting down the simulator interface.

        Subclasses should call the parent implementation, which will catch this
        method being called twice on the same `Simulator`.
        """
        if self._destroyed:
            raise RuntimeError("Simulator.destroy() called twice")
        self._destroyed = True


class Simulation(abc.ABC):
    """A single simulation run.

    These objects are not manipulated manually, but are created by a `Simulator`.
    Simulator interfaces should subclass this class, implementing various abstract
    methods to call the appropriate simulator APIs. In particular, the following
    methods must be implemented:

        * `createObjectInSimulator`, to create an object;
        * `step`, to run the simulation for one time step;
        * `getProperties`, to read back the new state of an object.

    Other methods can be overridden if necessary, e.g. `setup` for initialization
    at the start of the simulation and `destroy` for cleanup afterward.

    .. versionchanged:: 3.0

        The ``__init__`` method of subclasses should no longer create objects;
        the `createObjectInSimulator` method will be called instead. Other
        initialization which needs to take place after object creation should be
        done in `setup` after calling the superclass implementation.

        The arguments to ``__init__`` are the same as those to `simulate`, except
        that ``maxIterations`` is omitted.

    Attributes:
        currentTime (int): Number of time steps elapsed so far.
        timestep (float): Length of each time step in seconds.
        objects: List of Scenic objects (instances of `Object`) existing in the
            simulation. This list will change if objects are created dynamically.
        agents: List of :term:`agents` in the simulation. An agent is any object that has
            or had a behavior at any point in the simulation. The agents list may have objects
            appended to the end as the simulation progresses (if a non-agent object has its
            behavior overridden), but once an object is in the agents list its position is fixed.
        result (`SimulationResult`): Result of the simulation, or `None` if it has not
            yet completed. This is the primary object which should be inspected to get
            data out of the simulation: the other undocumented attributes of this class
            are for internal use only.

    Raises:
        RejectSimulationException: if a requirement is violated.
    """

    def __init_subclass__(cls):
        super().__init_subclass__()

        if hasattr(cls, "run") or hasattr(cls, "createObject"):
            raise RuntimeError(
                f"{cls.__name__} implements old-style simulation API; "
                "see documentation for how to upgrade to support Scenic 3"
            )

    def __init__(
        self,
        scene,
        *,
        maxSteps,
        name,
        timestep,
        replay=None,
        enableReplay=True,
        allowPickle=False,
        enableDivergenceCheck=False,
        divergenceTolerance=0,
        continueAfterDivergence=False,
        verbosity=0,
    ):
        self.result = None
        self.scene = scene
        self.objects = []
        self.trajectory = []
        self.records = defaultdict(list)
        self.currentTime = 0
        self.timestep = 1 if timestep is None else float(timestep)
        self.verbosity = verbosity
        self.name = name
        self.worker_num = 0

        self.actionSequence = []

        # Prepare to save or load a replay.
        self.initializeReplay(replay, enableReplay, enableDivergenceCheck, allowPickle)
        self.divergenceTolerance = divergenceTolerance
        self.continueAfterDivergence = continueAfterDivergence

        # Do the actual setup and execution of the simulation inside a try-finally
        # statement so that we roll back global state even if an error occurs.
        try:
            # Prepare global veneer state for running the simulation.
            import scenic.syntax.veneer as veneer

            veneer.beginSimulation(self)
            dynamicScenario = self.scene.dynamicScenario

            # Create objects and perform simulator-specific initialization.
            self.setup()

            # Initialize the top-level dynamic scenario.
            dynamicScenario._start()

            # Update all objects in case the simulator has adjusted any dynamic
            # properties during setup.
            self.updateObjects()

            # Run the simulation.
            terminationType, terminationReason = self._run(dynamicScenario, maxSteps)

            # Stop all remaining scenarios.
            # (and reject if some 'require eventually' condition was never satisfied)
            for scenario in tuple(reversed(veneer.runningScenarios)):
                scenario._stop("simulation terminated")

            # Record finally-recorded values.
            values = dynamicScenario._evaluateRecordedExprs(RequirementType.recordFinal)
            for name, val in values.items():
                self.records[name] = val

            # Package up simulation results into a compact object.
            result = SimulationResult(
                self.trajectory,
                self.actionSequence,
                terminationType,
                terminationReason,
                self.records,
            )
            self.result = result
        except (RejectSimulationException, RejectionException, GuardViolation) as e:
            # This simulation will be thrown out, but attach it to the exception
            # to aid in debugging.
            e.simulation = self
            raise
        finally:
            self.destroy()
            for obj in self.objects:
                disableDynamicProxyFor(obj)
            for agent in self.agents:
                if agent.behavior and agent.behavior._isRunning:
                    agent.behavior._stop()
            # If the simulation was terminated by an exception (including rejections),
            # some scenarios may still be running; we need to clean them up without
            # checking their requirements, which could raise rejection exceptions.
            for scenario in tuple(reversed(veneer.runningScenarios)):
                scenario._stop("exception", quiet=True)
            veneer.endSimulation(self)

    def _run(self, dynamicScenario, maxSteps):
        assert self.currentTime == 0

        while True:
            if self.verbosity >= 3:
                print(f"    Time step {self.currentTime}:")

            # Run compose blocks of compositional scenarios
            # (and check if any requirements defined therein fail)
            # N.B. if the top-level scenario completes, we don't immediately end
            # the simulation since we need to check if any monitors reject first.
            terminationReason = dynamicScenario._step()
            terminationType = TerminationType.scenarioComplete

            # Record current state of the simulation
            self.recordCurrentState()

            # Run monitors
            newReason = dynamicScenario._runMonitors()
            if newReason is not None:
                terminationReason = newReason
                terminationType = TerminationType.terminatedByMonitor

            # "Always" and scenario-level requirements have been checked;
            # now safe to terminate if the top-level scenario has finished,
            # a monitor requested termination, or we've hit the timeout
            if terminationReason is not None:
                return terminationType, terminationReason
            terminationReason = dynamicScenario._checkSimulationTerminationConditions()
            if terminationReason is not None:
                return TerminationType.simulationTerminationCondition, terminationReason
            if maxSteps and self.currentTime >= maxSteps:
                return TerminationType.timeLimit, f"reached time limit ({maxSteps} steps)"

            # Clear lastActions for all objects
            for obj in self.objects:
                obj.lastActions = tuple()

            # Update agents with any objects that now have behaviors (and are not already agents)
            self.agents += [
                obj for obj in self.objects if obj.behavior and obj not in self.agents
            ]

            # Compute the actions of the agents in this time step
            allActions = defaultdict(tuple)
            schedule = self.scheduleForAgents()
            if not set(self.agents) == set(schedule):
                raise RuntimeError("Simulator schedule does not contain all agents")
            for agent in schedule:
                # If agent doesn't have a behavior right now, continue
                if not agent.behavior:
                    continue

                # Run the agent's behavior to get its actions
                actions = agent.behavior._step()

                # Handle pseudo-actions marking the end of a simulation/scenario
                if isinstance(actions, _EndSimulationAction):
                    return TerminationType.terminatedByBehavior, str(actions)
                elif isinstance(actions, _EndScenarioAction):
                    scenario = actions.scenario
                    if scenario._isRunning:
                        scenario._stop(actions)
                    if scenario is dynamicScenario:
                        # Top-level scenario was terminated, so whole simulation will end.
                        return TerminationType.terminatedByBehavior, str(actions)
                    actions = ()

                # Check ordinary actions for compatibility
                assert isinstance(actions, tuple)
                if len(actions) == 1 and isinstance(actions[0], (list, tuple)):
                    actions = tuple(actions[0])
                if not self.actionsAreCompatible(agent, actions):
                    raise InvalidScenarioError(
                        f"agent {agent} tried incompatible action(s) {actions}"
                    )

                # Save actions for execution below
                allActions[agent] = actions

                # Log lastActions
                agent.lastActions = actions

            # Execute the actions
            if self.verbosity >= 3:
                for agent, actions in allActions.items():
                    print(f"      Agent {agent} takes action(s) {actions}")
            self.actionSequence.append(allActions)
            self.executeActions(allActions)

            # Run the simulation for a single step and read its state back into Scenic
            self.step()
            self.currentTime += 1
            self.updateObjects()

    def setup(self):
        """Set up the simulation to run in the simulator.

        Subclasses may override this method to perform custom initialization,
        but should call the parent implementation to create the objects in the
        initial scene (through `createObjectInSimulator`).
        """
        self.agents = []
        for obj in self.scene.objects:
            self._createObject(obj)

    def initializeReplay(self, replay, enableReplay, enableDivergenceCheck, allowPickle):
        if replay:
            self.replaying = True
            self._replayIn = Serializer(replay, allowPickle=allowPickle, detectEnd=True)
            flags = ReplayMode(self._replayIn.readReplayHeader())
            self._checkDivergence = ReplayMode.checkDivergence in flags
        else:
            self.replaying = False
        if enableReplay:
            self._replayOut = Serializer(allowPickle=allowPickle)
            flags = 0
            if enableDivergenceCheck:
                flags |= ReplayMode.checkDivergence
                self._writeDivergenceData = True
            else:
                self._writeDivergenceData = False
            self._replayOut.writeReplayHeader(flags)
        else:
            self._replayOut = None

    def _createObject(self, obj):
        if self.verbosity >= 3:
            print(f"      Creating object {obj}")

        # Add the new object to our lists.
        self.objects.append(obj)
        if obj.behavior:
            self.agents.append(obj)

        # Enable dynamic proxy for the object so that any mutations will not
        # affect the original object (e.g. if the simulator sets some of its
        # properties below).
        enableDynamicProxyFor(obj)

        # Ask subclass to actually create the object in the simulator.
        # (This may fail by raising an exception.)
        self.createObjectInSimulator(obj)

        # Allow the object to do simulator-specific initialization.
        obj.startDynamicSimulation()

    @abc.abstractmethod
    def createObjectInSimulator(self, obj):
        """Create the given object in the simulator.

        Implemented by subclasses. Should raise `SimulationCreationError` if creating
        the object fails.

        Args:
            obj (Object): the Scenic object to create.

        Raises:
            SimulationCreationError: if unable to create the object in the simulator.
        """
        raise NotImplementedError

    def recordCurrentState(self):
        dynamicScenario = self.scene.dynamicScenario
        records = self.records

        # Record initially-recorded values
        if self.currentTime == 0:
            values = dynamicScenario._evaluateRecordedExprs(RequirementType.recordInitial)
            for name, val in values.items():
                records[name] = val

        # Record time-series values
        values = dynamicScenario._evaluateRecordedExprs(RequirementType.record)
        for name, val in values.items():
            records[name].append((self.currentTime, val))

        self.trajectory.append(self.currentState())

    def replayCanContinue(self):
        if not self.replaying:
            return False
        self.detectReplayEnd()  # updates self.replaying
        return self.replaying

    def detectReplayEnd(self):
        if not self._replayIn.atEnd():
            return
        self.replaying = False
        if self.verbosity >= 2:
            print(f"    Continuing past end of replay at time step {self.currentTime}")

    def recordSampledValue(self, dist, values):
        if self._replayOut:
            dist.serializeValue(values, self._replayOut)

    def replaySampledValue(self, dist, values):
        return dist.deserializeValue(self._replayIn, values)

    def scheduleForAgents(self):
        """Compute the order for the agents to run in the next time step.

        The default order is the order in which the agents were created.

        Returns:
            An :term:`iterable` which is a permutation of ``self.agents``.
        """
        return self.agents

    def actionsAreCompatible(self, agent, actions):
        """Check whether the given actions can be taken simultaneously by an agent.

        The default is to consider all actions compatible with each other, and to
        call `Action.canBeTakenBy` to determine if an agent can take an action.
        Subclasses should override this method as appropriate.

        Args:
            agent (Object): the agent which wants to take the given actions.
            actions (tuple): tuple of :term:`actions` to be taken.
        """
        for action in actions:
            if not action.canBeTakenBy(agent):
                return False
        return True

    def executeActions(self, allActions):
        """Execute the actions selected by the agents.

        The default implementation calls the `applyTo` method of each `Action` to apply
        it to the appropriate agent.
        Subclasses may override this method to make additional simulator API calls as
        needed, but should call this implementation too or otherwise emulate its
        functionality.

        Args:
            allActions: a :obj:`~collections.defaultdict` mapping each agent to a tuple
                of actions, with the default value being an empty tuple. The order of
                agents in the dict should be respected in case the order of actions matters.
        """
        for agent, actions in allActions.items():
            for action in actions:
                action.applyTo(agent, self)

    @abc.abstractmethod
    def step(self):
        """Run the simulation for one step and return the next trajectory element.

        Implemented by subclasses. This should cause the simulator to simulate physics
        for ``self.timestep`` seconds.
        """
        raise NotImplementedError

    def updateObjects(self):
        """Update the positions and other properties of objects from the simulation.

        Subclasses likely do not need to override this method: they should implement its
        subroutine `getProperties` below.
        """
        for obj in self.objects:
            # Get latest values of dynamic properties from simulation and assign them
            dynTypes = obj._simulatorProvidedProperties
            properties = set(dynTypes)
            values = self.getProperties(obj, properties)
            assert properties == set(values), properties ^ set(values)
            for prop, value in values.items():
                # Check new value has the expected type
                ty = dynTypes[prop]
                if ty is float and isinstance(value, numbers.Real):
                    # Special case for scalars so that we don't penalize simulator interfaces
                    # for returning ints, NumPy scalar types, etc.
                    value = float(value)
                elif ty is type(None):
                    # Special case for properties with initial value None: the simulator sets
                    # their actual initial value, so we'll assume the type is correct here.
                    ty = type(value)
                    dynTypes[prop] = ty
                if not isinstance(value, ty):
                    actual = type(value).__name__
                    expected = ty.__name__
                    raise RuntimeError(
                        f'simulator provided value for property "{prop}" '
                        f"with type {actual} instead of expected {expected}"
                    )

                # Assign the new value
                setattr(obj, prop, value)

            # If saving a replay with divergence-checking support, save all the new values;
            # if running a replay with such support, check for divergence.
            if self._replayOut and self._writeDivergenceData:
                for prop, ty in dynTypes.items():
                    self._replayOut.writeValue(values[prop], ty)
            if self.replayCanContinue() and self._checkDivergence:
                for prop, ty in dynTypes.items():
                    expected = self._replayIn.readValue(ty)
                    actual = values[prop]
                    if self.valuesHaveDiverged(obj, prop, expected, actual):
                        msg = (
                            f'expected "{prop}" of {obj} to have value '
                            f"{expected}, but got {actual}"
                        )
                        if self.continueAfterDivergence:
                            if self.verbosity >= 2:
                                print(
                                    f"    Continuing past divergence at time step "
                                    f"{self.currentTime} ({msg})"
                                )
                            self.replaying = False
                            break
                        else:
                            raise DivergenceError(msg)

            # Recompute dynamic final properties
            obj._recomputeDynamicFinals()

            # Clear caches to ensure that cached properties like visibleRegion, etc.
            # are recomputed
            obj._clearCaches()

    def valuesHaveDiverged(self, obj, prop, expected, actual):
        """Decide whether the value of a dynamic property has diverged from the replay.

        The default implementation considers scalar and vector properties to have diverged
        if the distance between the actual and expected values is greater than
        ``self.divergenceTolerance`` (which is 0 by default); other types of properties
        use the != operator.

        Subclasses may override this function to provide more specialized criteria (e.g.
        allowing some properties to diverge more than others).

        Args:
            obj (Object): The object being considered.
            prop (str): The name of the :term:`dynamic property` being considered.
            expected: The value of the property saved in the replay currently being run.
            actual: The value of the property in the current simulation.

        Returns:
            `True` if the actual value should be considered as having diverged from the
            expected one; otherwise `False`.
        """
        diff = None
        if isinstance(expected, numbers.Real):
            diff = actual - expected
        elif isinstance(expected, Vector):
            diff = (actual - expected).norm()
        if diff:
            return diff > self.divergenceTolerance
        else:
            return actual != expected

    @abc.abstractmethod
    def getProperties(self, obj, properties):
        """Read the values of the given properties of the object from the simulator.

        Implemented by subclasses.

        Args:
            obj (Object): Scenic object in question.
            properties (set): Set of names of properties to read from the simulator.
                It is safe to destructively iterate through the set if you want.

        Returns:
            A `dict` mapping each of the given properties to its current value.
        """
        raise NotImplementedError

    def currentState(self):
        """Return the current state of the simulation.

        The definition of 'state' is up to the simulator; the 'state' is simply saved
        at each time step to define the 'trajectory' of the simulation.

        The default implementation returns a tuple of the positions of all objects.
        """
        return tuple(obj.position for obj in self.objects)

    @property
    def currentRealTime(self):
        """Current simulation time, in seconds."""
        return self.currentTime * self.timestep

    def destroy(self):
        """Perform any cleanup necessary to reset the simulator after a simulation.

        The default implementation does nothing by default; it may be overridden
        by subclasses.
        """
        pass

    def getReplay(self):
        """Encode this simulation to a `bytes` object for future replay.

        Requires that the simulation was run with ``enableReplay=True`` (the default).
        """
        if not self._replayOut:
            raise RuntimeError("cannot save replay without replay support enabled")
        return self._replayOut.getBytes()


class ReplayMode(enum.IntFlag):
    checkDivergence = enum.auto()


class DummySimulator(Simulator):
    """Simulator which does (almost) nothing, for testing and debugging purposes.

    To allow testing the change of dynamic properties over time, all objects drift
    upward by **drift** every time step.
    """

    def __init__(self, drift=0):
        super().__init__()
        self.drift = drift

    def createSimulation(self, scene, **kwargs):
        return DummySimulation(scene, drift=self.drift, **kwargs)


class DummySimulation(Simulation):
    """Minimal `Simulation` subclass for `DummySimulator`."""

    def __init__(self, scene, drift=0, **kwargs):
        self.drift = drift
        super().__init__(scene, **kwargs)

    def createObjectInSimulator(self, obj):
        pass

    def actionsAreCompatible(self, agent, actions):
        return True  # Allow non-Action actions for testing purposes

    def executeActions(self, allActions):
        pass

    def step(self):
        for obj in self.objects:
            obj.position += Vector(0, self.drift)

    def getProperties(self, obj, properties):
        vals = dict(
            position=obj.position,
            yaw=obj.yaw,
            pitch=obj.pitch,
            roll=obj.roll,
            velocity=Vector(0, 0, 0),
            angularVelocity=Vector(0, 0, 0),
            speed=0.0,
            angularSpeed=0.0,
        )
        for prop in properties:
            if prop not in vals:
                vals[prop] = None
        return vals


@enum.unique
class TerminationType(enum.Enum):
    """Enum describing the possible ways a simulation can end."""

    #: Simulation reached the specified time limit.
    timeLimit = "reached simulation time limit"

    #: The top-level scenario finished executing.
    #:
    #: (Either its :keyword:`compose` block completed, one of its termination
    #: conditions was met, or it was terminated with :keyword:`terminate`.)
    scenarioComplete = "the top-level scenario finished"

    #: A user-specified simulation termination condition was met.
    simulationTerminationCondition = "a simulation termination condition was met"

    #: A :term:`monitor` used :keyword:`terminate simulation` to end the simulation.
    terminatedByMonitor = "a monitor terminated the simulation"

    #: A :term:`dynamic behavior` used :keyword:`terminate simulation` to end the simulation.
    terminatedByBehavior = "a behavior terminated the simulation"


class SimulationResult:
    """Result of running a simulation.

    Attributes:
        trajectory: A tuple giving for each time step the simulation's 'state': by
            default the positions of every object. See `Simulation.currentState`.
        finalState: The last 'state' of the simulation, as above.
        actions: A tuple giving for each time step a dict specifying for each agent the
            (possibly-empty) tuple of actions it took at that time step.
        terminationType (`TerminationType`): The way the simulation ended.
        terminationReason (str): A human-readable string giving the reason why the
            simulation ended, possibly including debugging info.
        records (dict): For each :keyword:`record` statement, the value or time series of
            values its expression took during the simulation.
    """

    def __init__(self, trajectory, actions, terminationType, terminationReason, records):
        self.trajectory = tuple(trajectory)
        assert self.trajectory
        self.finalState = self.trajectory[-1]
        self.actions = tuple(actions)
        self.terminationType = terminationType
        self.terminationReason = str(terminationReason)
        self.records = dict(records)



"""Common exceptions and error handling."""

import importlib
import itertools
import pathlib
import pdb
import sys
import traceback
import types
import warnings

import scenic
import scenic.core
import scenic.syntax

## Configuration


def setDebuggingOptions(
    *, verbosity=0, fullBacktrace=False, debugExceptions=False, debugRejections=False
):
    """Configure Scenic's debugging options.

    Args:
        verbosity (int): Verbosity level. Zero by default, although the command-line
            interface uses 1 by default. See the :option:`--verbosity` option for the
            allowed values.
        fullBacktrace (bool): Whether to include Scenic's innards in backtraces
            (like the :option:`-b` command-line option).
        debugExceptions (bool): Whether to use `pdb` for post-mortem debugging of
            uncaught exceptions (like the :option:`--pdb` option).
        debugRejections (bool): Whether to enter `pdb` when a scene or simulation is
            rejected (like the :option:`--pdb-on-reject` option).
    """
    global verbosityLevel, showInternalBacktrace, postMortemDebugging
    global postMortemRejections
    verbosityLevel = verbosity
    postMortemDebugging = debugExceptions
    postMortemRejections = debugRejections
    if debugExceptions or debugRejections:
        fullBacktrace = True  # partial backtraces would be confusing in the debugger
    showInternalBacktrace = fullBacktrace


#: Verbosity level. See :option:`--verbosity` for the allowed values.
verbosityLevel = 0

#: Whether or not to include Scenic's innards in backtraces.
#:
#: Set to True by default so that any errors during import of the scenic module
#: will get full backtraces; the :mod:`scenic` module's *__init__.py* sets it to False.
showInternalBacktrace = True

#: Whether or not to do post-mortem debugging of uncaught exceptions.
postMortemDebugging = False

#: Whether or not to do "post-mortem" debugging of rejected scenes/simulations.
postMortemRejections = False

# fmt: off
#: Folders elided from backtraces when :obj:`showInternalBacktrace` is false.
#:
#: :meta hide-value:
hiddenFolders = [
    pathlib.Path(scenic.core.__file__).parent,      # scenic.core submodules
    pathlib.Path(scenic.syntax.__file__).parent,    # scenic.syntax submodules
    '<frozen importlib._bootstrap>',                # parts of importlib used internally
    pathlib.Path(importlib.__file__).parent,
]
# fmt: on

## Exceptions


class ScenicError(Exception):
    """An error produced during Scenic compilation, scene generation, or simulation."""

    pass


class ScenicSyntaxError(ScenicError):
    """An error produced by attempting to parse an invalid Scenic program.

    This is intentionally not a subclass of SyntaxError so that pdb can be used
    for post-mortem debugging of the parser. Our custom excepthook below will
    arrange to still have it formatted as a SyntaxError, though.
    """

    pass


class ParseCompileError(ScenicSyntaxError):
    """Error occurring during Scenic/Python parsing or compilation."""

    def __init__(self, exc):
        self.msg = exc.args[0]
        self.filename, self.lineno = exc.filename, exc.lineno
        self.end_lineno = getattr(exc, "end_lineno", None)  # added in Python 3.10
        end_offset = getattr(exc, "end_offset", None)  # ditto
        self.text, self.offset, self.end_offset = getText(
            self.filename, self.lineno, exc.text, exc.offset, end_offset
        )
        super().__init__(self.msg)
        self.with_traceback(exc.__traceback__)


class ScenicParseError(ParseCompileError):
    """Error occuring during Scenic parsing or compilation."""

    pass


class PythonCompileError(ParseCompileError):
    """Error occuring during Python compilation of translated Scenic code."""

    pass


class ASTParseError(ScenicSyntaxError):
    """Parse error occuring during modification of the Python AST."""

    def __init__(self, node, message, filename):
        self.msg = message
        self.lineno = node.lineno
        self.end_lineno = getattr(node, "end_lineno", None)  # not always present
        end_offset = getattr(node, "end_col_offset", None)  # ditto
        self.filename = filename
        self.text, self.offset, self.end_offset = getText(
            filename, node.lineno, offset=node.col_offset, end_offset=end_offset
        )
        super().__init__(message)


class InvalidScenarioError(ScenicError):
    """Error raised for syntactically-valid but otherwise problematic Scenic programs."""

    pass


class InconsistentScenarioError(InvalidScenarioError):
    """Error for scenarios with inconsistent requirements."""

    def __init__(self, line, message):
        self.lineno = line
        super().__init__("Inconsistent requirement on line " + str(line) + ": " + message)


class SpecifierError(ScenicError):
    """Error for illegal uses of specifiers."""

    pass


## Scenic backtraces


def excepthook(ty, value, tb):
    if sys.version_info >= (3, 11) and issubclass(ty, BaseExceptionGroup):
        # Use the default mechanism since we don't handle ExceptionGroups
        sys.__excepthook__(ty, value, tb)
    else:
        displayScenicException(value)

    if postMortemDebugging:
        print("Uncaught exception. Entering post-mortem debugging")
        import pdb

        pdb.post_mortem(tb)


def displayScenicException(exc, seen=None):
    """Print a Scenic exception, cleaning up the traceback if desired.

    If ``showInternalBacktrace`` is False, this hides frames inside Scenic itself.
    """
    ty = type(exc)
    tb = exc.__traceback__

    # Recursively print the cause/context for this exception, if any.
    # (This code is adapted from IDLE.)
    if seen is None:
        seen = set()
    seen.add(id(exc))
    cause = exc.__cause__
    context = exc.__context__
    if cause is not None and id(cause) not in seen:
        displayScenicException(cause, seen)
        print(
            "\nThe above exception was the direct cause of the following exception:\n",
            file=sys.stderr,
        )
    elif context is not None and not exc.__suppress_context__ and id(context) not in seen:
        displayScenicException(context, seen)
        print(
            "\nDuring handling of the above exception, another exception occurred:\n",
            file=sys.stderr,
        )

    if showInternalBacktrace:
        strings = ["Traceback (most recent call last):\n"]
    else:
        strings = [
            "Traceback (most recent call last; use -b to show Scenic internals):\n"
        ]

    # Work out how to present the exception type
    pseudoSyntaxError = issubclass(ty, ScenicSyntaxError)
    if issubclass(ty, ScenicError):
        if issubclass(ty, ScenicSyntaxError) and not showInternalBacktrace:
            name = "ScenicSyntaxError"
        else:
            name = ty.__name__
        if pseudoSyntaxError:
            # hack to get format_exception_only to format this like a bona fide SyntaxError
            bases = (SyntaxError,)
        else:
            bases = ty.__bases__
        formatTy = type(name, bases, {})
        if not showInternalBacktrace:
            formatTy.__module__ = "__main__"  # hide qualified name of the exception
    else:
        formatTy = ty

    if issubclass(ty, SyntaxError) or (pseudoSyntaxError and not showInternalBacktrace):
        pass  # no backtrace for these types of errors
    elif loc := getattr(exc, "_scenic_location", None):
        strings.extend(traceback.format_list([loc]))
    else:
        summary = traceback.extract_tb(tb)
        if showInternalBacktrace:
            filtered = summary
        else:
            filtered = []
            skip = False
            for frame in summary:
                if skip:
                    skip = False
                    continue
                if frame.name == "callBeginningScenicTrace":
                    filtered = []
                    skip = True
                elif includeFrame(frame):
                    filtered.append(frame)
        if filtered:
            strings.extend(traceback.format_list(filtered))
        else:
            strings.append("  <Scenic internals>\n")
    # Note: we can't directly call traceback.format_exception_only any more,
    # since as of Python 3.10 it ignores the exception class passed in and
    # uses type(exc) instead, foiling our formatTy hack.
    tbe = traceback.TracebackException(formatTy, exc, None)
    strings.extend(tbe.format_exception_only())
    message = "".join(strings)
    print(message, end="", file=sys.stderr)


def includeFrame(frame):
    if frame.filename in hiddenFolders:
        return False
    parents = pathlib.Path(frame.filename).parents
    return not any(folder in parents for folder in hiddenFolders)


# Install our excepthook if nobody else has already installed one;
# we specially allow overwriting excepthooks from the following modules,
# which are known to not cause problems:
#   - exceptiongroup (PEP 654 backport for pre-3.11)
if sys.excepthook is not sys.__excepthook__ and not sys.excepthook.__module__.startswith(
    "exceptiongroup"
):
    warnings.warn("unable to install sys.excepthook to format Scenic backtraces")
else:
    sys.excepthook = excepthook


def callBeginningScenicTrace(func):
    """Call the given function, starting the Scenic backtrace at that point.

    This function is just a convenience to make Scenic backtraces cleaner when
    running Scenic programs from the command line.
    """
    return func()


def saveErrorLocation():
    stack = traceback.extract_stack()
    for frame in reversed(stack):
        if includeFrame(frame):
            return frame
    return None


## Utilities


def optionallyDebugRejection(exc=None):
    if not postMortemRejections:
        return
    print("Scene/simulation rejected. Entering debugger...")
    if exc:
        pdb.post_mortem(exc.__traceback__)
    else:
        debugger = pdb.Pdb()
        debugger.set_trace(sys._getframe().f_back)  # TODO more portable way to do this?


def getText(filename, lineno, line="", offset=0, end_offset=None):
    """Attempt to recover the text of an error from the original Scenic file."""
    if not filename or filename in {"<stream>", "<string>", "<unknown>"}:
        return line, offset, end_offset  # don't bother with the filesystem
    try:
        with open(filename, "r") as f:
            line = list(itertools.islice(f, lineno - 1, lineno))
        line = line[0] if line else ""
        # TODO improve reconstruction of error position?
        # (see TracebackException._format_syntax_error for spaces calculation below)
        rline = line.rstrip("\n")
        lline = rline.lstrip(" \n\f")
        spaces = len(rline) - len(lline)
        if end_offset is not None:
            adj_end_offset = min(end_offset, len(line) + 1)
            offset = max(spaces + 1, adj_end_offset - (end_offset - offset))
            end_offset = adj_end_offset
        else:
            offset = min(offset, len(line))
    except OSError:
        pass
    return line, offset, end_offset



"""Dynamic scenarios."""

import ast
from collections import defaultdict
import dataclasses
import functools
import inspect
import weakref

import rv_ltl

import scenic
import scenic.core.dynamics as dynamics
from scenic.core.errors import InvalidScenarioError, ScenicSyntaxError
from scenic.core.lazy_eval import DelayedArgument, needsLazyEvaluation
from scenic.core.requirements import (
    DynamicRequirement,
    PendingRequirement,
    RequirementType,
)
from scenic.core.utils import alarm, argsToString
from scenic.core.workspaces import Workspace

from .actions import _EndScenarioAction, _EndSimulationAction
from .behaviors import Behavior, Monitor
from .invocables import Invocable
from .utils import RejectSimulationException, StuckBehaviorWarning


class DynamicScenario(Invocable):
    """Internal class for scenarios which can execute during dynamic simulations.

    Provides additional information complementing `Scenario`, which originally only
    supported static scenarios. The two classes should probably eventually be merged.
    """

    def __init_subclass__(cls, *args, **kwargs):
        import scenic.syntax.veneer as veneer

        veneer.registerDynamicScenarioClass(cls)

        target = cls._setup or cls._compose or (lambda self, agent: 0)
        target = functools.partial(target, 0, 0)  # account for Scenic-inserted args
        cls.__signature__ = inspect.signature(target)

    _requirementSyntax = None  # overridden by subclasses
    _simulatorFactory = None
    _globalParameters = None
    _locals = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ego = None
        self._workspace = None
        self._instances = []  # ordered for reproducibility
        # _objects should contain a reference to the most complete version of
        # the objects in this scene (sampled > unsampled)
        self._objects = []  # ordered for reproducibility
        self._sampledObjects = self._objects
        self._externalParameters = []
        self._pendingRequirements = defaultdict(list)
        self._requirements = []
        # things needing to be sampled to evaluate the requirements
        self._requirementDeps = set()

        self._agents = []
        self._monitors = []
        self._behaviors = []
        self._monitorRequirements = []
        self._temporalRequirements = []
        self._terminationConditions = []
        self._terminateSimulationConditions = []
        self._recordedExprs = []
        self._recordedInitialExprs = []
        self._recordedFinalExprs = []

        self._subScenarios = []
        self._endWithBehaviors = False
        self._timeLimit = None
        self._timeLimitIsInSeconds = False
        self._prepared = False
        self._delayingPreconditionCheck = False
        self._dummyNamespace = None

        self._timeLimitInSteps = None  # computed at simulation time
        self._elapsedTime = 0
        self._eventuallySatisfied = None
        self._overrides = {}

        self._requirementMonitors = None

    @classmethod
    def _dummy(cls, namespace):
        scenario = cls()
        scenario._setup = None
        scenario._compose = None
        scenario._prepared = True
        scenario._dummyNamespace = namespace
        return scenario

    @classmethod
    def _requiresArguments(cls):
        """Whether this scenario cannot be instantiated without arguments."""
        if cls._setup:
            func = cls._setup
        elif cls._compose:
            func = cls._compose
        else:
            return True
        sig = inspect.signature(func)
        try:
            sig.bind(None, None)  # first two arguments are added internally by Scenic
            return False
        except TypeError:
            return True

    @property
    def ego(self):
        if self._ego is None:
            return DelayedArgument((), lambda context: self._ego, _internal=True)
        return self._ego

    @property
    def objects(self):
        return tuple(self._objects)

    def _bindTo(self, scene):
        """Bind this scenario to a sampled scene when starting a new simulation."""
        self._ego = scene.egoObject
        self._workspace = scene.workspace
        self._objects = list(scene.objects)
        self._agents = [obj for obj in scene.objects if obj.behavior is not None]
        self._monitors = list(scene.monitors)
        self._temporalRequirements = scene.temporalRequirements
        self._terminationConditions = scene.terminationConditions
        self._terminateSimulationConditions = scene.terminateSimulationConditions
        self._recordedExprs = scene.recordedExprs
        self._recordedInitialExprs = scene.recordedInitialExprs
        self._recordedFinalExprs = scene.recordedFinalExprs

    def _prepare(self, delayPreconditionCheck=False):
        """Prepare the scenario for execution, executing its setup block."""
        import scenic.syntax.veneer as veneer

        assert not self._prepared
        self._prepared = True

        self._finalizeArguments()  # TODO generalize _prepare for Invocable?

        veneer.prepareScenario(self)
        with veneer.executeInScenario(self, inheritEgo=True):
            # Check preconditions and invariants
            if delayPreconditionCheck:
                self._delayingPreconditionCheck = True
            else:
                self._checkAllPreconditions()

            # Execute setup block
            if self._setup is not None:
                assert not any(needsLazyEvaluation(arg) for arg in self._args)
                assert not any(needsLazyEvaluation(arg) for arg in self._kwargs.values())
                self._setup(None, *self._args, **self._kwargs)
                veneer.finishScenarioSetup(self)

        # Extract requirements, scan for relations used for pruning, and create closures
        self._compileRequirements()

    @classmethod
    def _bindGlobals(cls, globs):
        cls._globalParameters = globs

    def _start(self):
        """Start the scenario, starting its compose block, behaviors, and monitors."""
        import scenic.syntax.veneer as veneer

        super()._start()
        assert self._prepared

        # Check preconditions if they could not be checked earlier
        if self._delayingPreconditionCheck:
            self._checkAllPreconditions()

        # Compute time limit now that we know the simulation timestep
        self._elapsedTime = 0
        self._timeLimitInSteps = self._timeLimit
        if self._timeLimitIsInSeconds:
            self._timeLimitInSteps /= veneer.currentSimulation.timestep

        # create monitors for each requirement used for this simulation
        self._requirementMonitors = [r.toMonitor() for r in self._temporalRequirements]

        veneer.startScenario(self)
        with veneer.executeInScenario(self):
            # Start compose block
            if self._compose is not None:
                if not inspect.isgeneratorfunction(self._compose):
                    from scenic.syntax.translator import composeBlock

                    raise InvalidScenarioError(
                        f'"{composeBlock}" does not invoke any scenarios'
                    )
                self._runningIterator = self._compose(None, *self._args, **self._kwargs)

            # Initialize behavior coroutines of agents
            for agent in self._agents:
                behavior = agent.behavior
                assert isinstance(behavior, Behavior), behavior
                behavior._assignTo(agent)
            # Initialize monitor coroutines
            for monitor in self._monitors:
                monitor._start()

    def _step(self):
        """Execute the (already-started) scenario for one time step.

        Returns:
            `None` if the scenario will continue executing; otherwise a string describing
            why it has terminated.
        """
        import scenic.syntax.veneer as veneer

        super()._step()

        # Check temporal requirements
        for m in self._requirementMonitors:
            result = m.value()
            if result == rv_ltl.B4.FALSE:
                raise RejectSimulationException(str(m))

        # Check if we have reached the time limit, if any
        if (
            self._timeLimitInSteps is not None
            and self._elapsedTime >= self._timeLimitInSteps
        ):
            return self._stop("reached time limit")
        self._elapsedTime += 1

        # Execute compose block, if any
        composeDone = False
        if self._runningIterator is None:
            composeDone = True  # compose block ended in an earlier step
        else:

            def alarmHandler(signum, frame):
                if sys.gettrace():
                    return  # skip the warning if we're in the debugger
                warnings.warn(
                    f"the compose block of scenario {self} is taking a long time; "
                    'maybe you have an infinite loop with no "wait" statement?',
                    StuckBehaviorWarning,
                )

            timeout = dynamics.stuckBehaviorWarningTimeout
            with veneer.executeInScenario(self), alarm(timeout, alarmHandler):
                try:
                    result = self._runningIterator.send(None)
                    if isinstance(result, (_EndSimulationAction, _EndScenarioAction)):
                        return self._stop(result)
                except StopIteration:
                    self._runningIterator = None
                    composeDone = True

        # If there is a compose block and it has finished, we're done
        if self._compose is not None and composeDone:
            return self._stop("finished compose block")

        # Optionally end when all our agents' behaviors have ended
        if self._endWithBehaviors:
            if all(agent.behavior._isFinished for agent in self._agents):
                return self._stop("all behaviors finished")

        # Check if any termination conditions apply
        for req in self._terminationConditions:
            if req.evaluate():
                return self._stop(req)

        # Scenario will not terminate yet
        return None

    def _stop(self, reason, quiet=False):
        """Stop the scenario's execution, for the given reason."""
        import scenic.syntax.veneer as veneer

        assert self._isRunning

        # Stop monitors and subscenarios.
        for monitor in self._monitors:
            if monitor._isRunning:
                monitor._stop()
        self._monitors = []
        for sub in self._subScenarios:
            if sub._isRunning:
                sub._stop("parent scenario ending", quiet=quiet)
        self._runningIterator = None

        # Revert overrides.
        for obj, oldVals in self._overrides.items():
            obj._revert(oldVals)

        # Inform the veneer we have stopped, and mark ourselves finished.
        veneer.endScenario(self, reason, quiet=quiet)
        super()._stop(reason)

        # Reject if a temporal requirement was not satisfied.
        if not quiet:
            for req in self._requirementMonitors:
                if req.lastValue.is_falsy:
                    raise RejectSimulationException(str(req))
        self._requirementMonitors = None

        return reason

    def _invokeInner(self, agent, subs):
        for sub in subs:
            if not isinstance(sub, DynamicScenario):
                raise TypeError(f"expected a scenario, got {sub}")
            sub._prepare()
            sub._start()
        self._subScenarios = list(subs)
        while True:
            newSubs = []
            for sub in self._subScenarios:
                terminationReason = sub._step()
                if isinstance(terminationReason, _EndSimulationAction):
                    yield terminationReason
                    assert False, self  # should never get here since simulation ends
                elif terminationReason is None:
                    newSubs.append(sub)
            self._subScenarios = newSubs
            if not newSubs:
                return
            yield None
            # Check if any sub-scenarios stopped during action execution
            self._subScenarios = [sub for sub in self._subScenarios if sub._isRunning]

    def _evaluateRecordedExprs(self, ty):
        if ty is RequirementType.record:
            place = "_recordedExprs"
        elif ty is RequirementType.recordInitial:
            place = "_recordedInitialExprs"
        elif ty is RequirementType.recordFinal:
            place = "_recordedFinalExprs"
        else:
            assert False, "invalid record type requested"
        return self._evaluateRecordedExprsAt(place)

    def _evaluateRecordedExprsAt(self, place):
        values = {}
        for rec in getattr(self, place):
            values[rec.name] = rec.evaluate()
        for sub in self._subScenarios:
            subvals = sub._evaluateRecordedExprsAt(place)
            values.update(subvals)
        return values

    def _runMonitors(self):
        terminationReason = None
        endScenario = None
        for monitor in self._monitors:
            action = monitor._step()
            # do not exit early, since subsequent monitors could reject the simulation
            if isinstance(action, _EndSimulationAction):
                terminationReason = action
            elif isinstance(action, _EndScenarioAction):
                assert action.scenario is None
                endScenario = action
        for sub in self._subScenarios:
            subreason = sub._runMonitors()
            if subreason is not None:
                terminationReason = subreason
        if endScenario:
            self._stop(endScenario)
        return terminationReason or endScenario

    def _checkSimulationTerminationConditions(self):
        for req in self._terminateSimulationConditions:
            if req.isTrue().is_truthy:
                return req
        return None

    @property
    def _allAgents(self):
        agents = list(self._agents)
        for sub in self._subScenarios:
            agents.extend(sub._allAgents)
        return agents

    def _inherit(self, other):
        if not self._workspace:
            self._workspace = other._workspace
        self._instances.extend(other._instances)
        self._objects.extend(other._objects)
        self._agents.extend(other._agents)
        self._globalParameters.update(other._globalParameters)
        self._externalParameters.extend(other._externalParameters)
        self._requirements.extend(other._requirements)
        self._behaviors.extend(other._behaviors)

    def _registerInstance(self, inst):
        self._instances.append(inst)

    def _registerObject(self, obj):
        self._registerInstance(obj)
        self._objects.append(obj)
        if getattr(obj, "behavior", None) is not None:
            self._agents.append(obj)

        obj._parentScenario = weakref.ref(self)

    def _addRequirement(self, ty, reqID, req, line, name, prob):
        """Save a requirement defined at compile-time for later processing."""
        assert reqID not in self._pendingRequirements
        preq = PendingRequirement(ty, req, line, prob, name, self._ego)
        self._pendingRequirements[reqID] = preq

    def _addDynamicRequirement(self, ty, req, line, name):
        """Add a requirement defined during a dynamic simulation."""
        dreq = DynamicRequirement(ty, req, line, name)
        self._temporalRequirements.append(dreq)

    def _addMonitor(self, monitor):
        """Add a monitor during a dynamic simulation."""
        assert isinstance(monitor, Monitor)
        self._monitors.append(monitor)
        if self._isRunning:
            monitor._start()

    def _compileRequirements(self):
        namespace = self._dummyNamespace if self._dummyNamespace else self.__dict__
        requirementSyntax = self._requirementSyntax
        assert requirementSyntax is not None
        for reqID, requirement in self._pendingRequirements.items():
            syntax = requirementSyntax[reqID] if requirementSyntax else None

            # Catch the simple case where someone has most likely forgotten the "monitor"
            # keyword.
            if (
                (not requirement.ty == RequirementType.monitor)
                and isinstance(syntax, ast.Call)
                and isinstance(syntax.func, ast.Name)
                and syntax.func.id in namespace
                and isinstance(namespace[syntax.func.id], type)
                and issubclass(
                    namespace[syntax.func.id], scenic.core.dynamics.behaviors.Monitor
                )
            ):
                raise ScenicSyntaxError(
                    f"Missing 'monitor' keyword after 'require' when instantiating '{syntax.func.id}'"
                )

            compiledReq = requirement.compile(namespace, self, syntax)

            self._registerCompiledRequirement(compiledReq)
            self._requirementDeps.update(compiledReq.dependencies)

    def _registerCompiledRequirement(self, req):
        if req.ty is RequirementType.require:
            place = self._requirements
        elif req.ty is RequirementType.monitor:
            place = self._monitorRequirements
        elif req.ty is RequirementType.terminateWhen:
            place = self._terminationConditions
        elif req.ty is RequirementType.terminateSimulationWhen:
            place = self._terminateSimulationConditions
        elif req.ty is RequirementType.record:
            place = self._recordedExprs
        elif req.ty is RequirementType.recordInitial:
            place = self._recordedInitialExprs
        elif req.ty is RequirementType.recordFinal:
            place = self._recordedFinalExprs
        else:
            raise RuntimeError(f"internal error: requirement {req} has unknown type!")
        place.append(req)

    def _setTimeLimit(self, timeLimit, inSeconds=True):
        self._timeLimit = timeLimit
        self._timeLimitIsInSeconds = inSeconds

    def _override(self, obj, specifiers):
        oldVals = obj._override(specifiers)
        if obj not in self._overrides:
            self._overrides[obj] = oldVals

    def _toScenario(self, namespace):
        assert self._prepared

        if not self._workspace:
            self._workspace = Workspace()  # default empty workspace
        astHash = namespace["_astHash"]
        name = None if self._dummyNamespace else self.__class__.__name__
        options = dataclasses.replace(namespace["_compileOptions"], scenario=name)

        from scenic.core.scenarios import Scenario

        scenario = Scenario(
            self._workspace,
            self._simulatorFactory,
            self._instances,
            self._objects,
            self._ego,
            self._globalParameters,
            self._externalParameters,
            self._requirements,
            self._requirementDeps,
            self._monitorRequirements,
            self._behaviorNamespaces,
            self,
            astHash,
            options,
        )  # TODO unify these!
        return scenario

    def __getattr__(self, name):
        if name in self._locals:
            return DelayedArgument(
                (), lambda context: getattr(self, name), _internal=True
            )
        return object.__getattribute__(self, name)

    def __str__(self):
        if self._dummyNamespace:
            return "top-level scenario"
        else:
            args = argsToString(self._args, self._kwargs)
            return f"{self.__class__.__name__}({args})"
